"""
A short Python script
=====================

This demonstrates an example ``.py`` file that is not executed when gallery is
generated (see :ref:`build_pattern`) but nevertheless gets included as an
example. Note that no output is capture as this file is not executed.
"""

# Code source: Óscar Nájera
# License: BSD 3 clause
print([i for i in range(10)])



"""
Failing example test with normal thumbnail behaviour
====================================================
Test files in expected_failing_examples run, but normal thumbnail behaviour is
retained when sphinx_gallery_failing_thumbnail = False.
"""

# sphinx_gallery_failing_thumbnail = False

import matplotlib.pyplot as plt
import numpy as np

plt.pcolormesh(np.random.randn(100, 100))

# %%
# will raise AssertionError

assert False



# Author: Óscar Nájera
# License: 3-clause BSD
"""Sphinx-Gallery Generator.

Attaches Sphinx-Gallery to Sphinx in order to generate the galleries
when building the documentation.
"""

import codecs
import copy
import os
import pathlib
import re
from datetime import datetime, timedelta
from difflib import get_close_matches
from itertools import chain
from pathlib import Path
from textwrap import indent
from xml.sax.saxutils import escape, quoteattr

import sphinx.util
from docutils import nodes
from sphinx.errors import ConfigError, ExtensionError
from sphinx.util.console import blue, bold, purple, red

from . import __version__ as _sg_version
from . import glr_path_static
from .backreferences import _finalize_backreferences
from .directives import ImageSg, MiniGallery, imagesg_addnode
from .docs_resolv import embed_code_links
from .downloads import generate_zipfiles
from .gen_rst import (
    SPHX_GLR_SIG,
    _get_call_memory_and_base,
    _get_callables,
    _get_gallery_header,
    generate_dir_rst,
)
from .interactive_example import (
    check_binder_conf,
    check_jupyterlite_conf,
    copy_binder_files,
    create_jupyterlite_contents,
    post_configure_jupyterlite_sphinx,
    pre_configure_jupyterlite_sphinx,
)
from .recommender import ExampleRecommender, _write_recommendations
from .scrapers import _import_matplotlib
from .sorting import ExplicitOrder
from .utils import (
    _collect_gallery_files,
    _combine_backreferences,
    _format_toctree,
    _has_graphviz,
    _has_optipng,
    _has_pypandoc,
    _read_json,
    _replace_md5,
    _write_json,
)

_KNOWN_CSS = (
    "sg_gallery",
    "sg_gallery-binder",
    "sg_gallery-dataframe",
    "sg_gallery-rendered-html",
)


class DefaultResetArgv:
    """Provides default 'reset_argv' callable that returns empty list."""

    def __repr__(self):
        return "DefaultResetArgv"

    def __call__(self, gallery_conf, script_vars):
        """Return empty list."""
        return []


DEFAULT_GALLERY_CONF = {
    "filename_pattern": re.escape(os.sep) + "plot",
    "ignore_pattern": r"__init__\.py",
    "examples_dirs": os.path.join("..", "examples"),
    "example_extensions": {".py"},
    "filetype_parsers": {},
    "notebook_extensions": {".py"},
    "reset_argv": DefaultResetArgv(),
    "subsection_order": None,
    "within_subsection_order": "NumberOfCodeLinesSortKey",
    "minigallery_sort_order": None,
    "gallery_dirs": "auto_examples",
    "backreferences_dir": None,
    "doc_module": (),
    "exclude_implicit_doc": set(),
    "reference_url": {},
    "capture_repr": ("_repr_html_", "__repr__"),
    "ignore_repr_types": r"",
    # Build options
    # -------------
    # 'plot_gallery' also accepts strings that evaluate to a bool, e.g. "True",
    # "False", "1", "0" so that they can be easily set via command line
    # switches of sphinx-build
    "plot_gallery": "True",
    "download_all_examples": True,
    "abort_on_example_error": False,
    "only_warn_on_example_error": False,
    "recommender": {"enable": False},
    "failing_examples": {},
    "passing_examples": [],
    "stale_examples": [],  # ones that did not need to be run due to md5sum
    "run_stale_examples": False,
    "expected_failing_examples": set(),
    "thumbnail_size": (400, 280),  # Default CSS does 0.4 scaling (160, 112)
    "min_reported_time": 0,
    "write_computation_times": os.getenv("SOURCE_DATE_EPOCH") is None,
    "binder": {},
    "jupyterlite": {},
    "promote_jupyter_magic": False,
    "image_scrapers": ("matplotlib",),
    "compress_images": (),
    "reset_modules": ("matplotlib", "seaborn"),
    "reset_modules_order": "before",
    "first_notebook_cell": None,
    "last_notebook_cell": None,
    "notebook_images": False,
    "pypandoc": False,
    "remove_config_comments": False,
    "show_memory": False,
    "show_signature": True,
    "junit": "",
    "log_level": {"backreference_missing": "warning"},
    "inspect_global_variables": True,
    "css": _KNOWN_CSS,
    "matplotlib_animations": False,
    "image_srcset": [],
    "default_thumb_file": None,
    "line_numbers": False,
    "nested_sections": True,
    "prefer_full_module": set(),
    "api_usage_ignore": ".*__.*__",
    "show_api_usage": False,  # if this changes, change write_api_entries, too
    "copyfile_regex": "",
    "parallel": False,
}

logger = sphinx.util.logging.getLogger("sphinx-gallery")


def _bool_eval(x):
    """Evaluate bool only configs, to allow setting via -D on the command line."""
    if isinstance(x, str):
        try:
            x = eval(x)
        except TypeError:
            pass
    return bool(x)


def _update_gallery_conf_exclude_implicit_doc(gallery_conf):
    """Update gallery config exclude_implicit_doc.

    This is separate function for better testability.
    """
    # prepare regex for exclusions from implicit documentation, ensuring that what
    # gets complied has a stable __repr__ (i.e., by sorting the exclude_implicit_doc
    # set before joining)
    exclude_regex = (
        re.compile("|".join(sorted(gallery_conf["exclude_implicit_doc"])))
        if gallery_conf["exclude_implicit_doc"]
        else False
    )
    gallery_conf["exclude_implicit_doc_regex"] = exclude_regex


def _update_gallery_conf_builder_inited(
    sphinx_gallery_conf,
    src_dir,
    plot_gallery=True,
    abort_on_example_error=False,
    builder_name="html",
):
    sphinx_gallery_conf.update(plot_gallery=plot_gallery)
    sphinx_gallery_conf.update(abort_on_example_error=abort_on_example_error)
    sphinx_gallery_conf["src_dir"] = src_dir
    # Make it easy to know which builder we're in
    sphinx_gallery_conf["builder_name"] = builder_name


def _check_extra_config_keys(gallery_conf, sphinx_gallery_conf, check_keys):
    """Check SG config keys, optionally raising if any extra keys present."""
    options = sorted(gallery_conf)
    extra_keys = sorted(set(sphinx_gallery_conf) - set(options))
    if extra_keys and check_keys:
        msg = "Unknown key(s) in sphinx_gallery_conf:\n"
        for key in extra_keys:
            options = get_close_matches(key, options, cutoff=0.66)
            msg += repr(key)
            if len(options) == 1:
                msg += f", did you mean {options[0]!r}?"
            elif len(options) > 1:
                msg += f", did you mean one of {options!r}?"
            msg += "\n"
        raise ConfigError(msg.strip())


def _check_config_type(
    gallery_conf,
    conf_key,
    types,
    str_to_list=False,
    allow_none=False,
):
    """Check config type, optionally converting str to list or allowing None."""
    conf_value = gallery_conf[conf_key]
    # Early exit if type correct
    if isinstance(conf_value, types) or (allow_none and conf_value is None):
        if str_to_list:
            if isinstance(conf_value, str):
                gallery_conf[conf_key] = [conf_value]
            return
        return
    # Otherwise raise error
    msg = "'{conf_key}' config allowed types: {types}{or_none}. Got {conf_type}."
    if isinstance(types, type):
        types = (types,)
    str_types = [t.__name__ for t in types]
    or_none = " or None" if allow_none else ""
    conf_type = type(conf_value).__name__
    msg = msg.format(
        conf_key=conf_key,
        types=str_types,
        or_none=or_none,
        conf_type=conf_type,
    )
    raise ConfigError(msg)


def _check_image_srcset(gallery_conf):
    """Check `_check_image_srcset`, convert to float and removing '1'."""
    _check_config_type(gallery_conf, "image_srcset", (list, tuple), str_to_list=True)
    srcset_mult_facs = set()
    for st in gallery_conf["image_srcset"]:
        if not (isinstance(st, str) and st[-1:] == "x"):
            raise ConfigError(
                f"Invalid value for image_srcset parameter: {st!r}. "
                "Must be a list of strings with the multiplicative "
                'factor followed by an "x".  e.g. ["2.0x", "1.5x"]'
            )
        # "2x" -> "2.0"
        srcset_mult_facs.add(float(st[:-1]))
    srcset_mult_facs -= {1}  # 1x is always saved.
    gallery_conf["image_srcset"] = [*sorted(srcset_mult_facs)]


def _check_compress_images(gallery_conf):
    """Check `compress_images`, getting any command line args."""
    _check_config_type(
        gallery_conf,
        "compress_images",
        (str, tuple, list),
        str_to_list=True,
    )
    compress_images = gallery_conf["compress_images"]
    compress_images = list(compress_images)
    allowed_values = ("images", "thumbnails")
    pops = list()
    # Get command-line switches
    for ki, kind in enumerate(compress_images):
        if kind not in allowed_values:
            if kind.startswith("-"):
                pops.append(ki)
                continue
            raise ConfigError(
                "All entries in compress_images must be one of "
                f"{allowed_values} or a command-line switch "
                f'starting with "-", got {kind!r}'
            )
    compress_images_args = [compress_images.pop(p) for p in pops[::-1]]
    if len(compress_images) and not _has_optipng():
        logger.warning(
            "optipng binaries not found, PNG %s will not be optimized",
            " and ".join(compress_images),
        )
        compress_images = ()
    gallery_conf["compress_images"] = compress_images
    gallery_conf["compress_images_args"] = compress_images_args


def _check_matplotlib_animations(gallery_conf, app):
    """Check `matplotlib_animations` config."""
    animations = gallery_conf["matplotlib_animations"]
    if isinstance(animations, bool):
        # We allow single boolean for backwards compatibility reasons
        animations = (animations,)
        fmt = None
    if not (len(animations) in (1, 2) and isinstance(enabled := animations[0], bool)):
        raise ConfigError(
            "'matplotlib_animations' must be a single bool or "
            f"(enabled: bool, format: str), not {animations!r}"
        )
    # Handle file format
    if len(animations) > 1:
        fmt = animations[1]
        if fmt is not None:
            if not isinstance(fmt, str):
                raise ConfigError(
                    "'matplotlib_animations' file format must be a string or None"
                )
            if fmt not in ("html5", "jshtml"):
                if app is not None:
                    # Other formats mean animations saved externally and require
                    # this `video` extension to embed them into the HTML
                    try:
                        app.setup_extension("sphinxcontrib.video")
                    except ExtensionError as e:
                        raise ConfigError(
                            f"'matplotlib_animations' specifies file format: {fmt}; "
                            f"this requires the sphinxcontrib.video package."
                        ) from e

    gallery_conf["matplotlib_animations"] = (enabled, fmt)


def _check_pypandoc_config(gallery_conf):
    """Check `pypandoc` config."""
    pypandoc = gallery_conf["pypandoc"]
    _check_config_type(gallery_conf, "pypandoc", (dict, bool))

    gallery_conf["pypandoc"] = dict() if pypandoc is True else pypandoc
    has_pypandoc, version = _has_pypandoc()
    if isinstance(gallery_conf["pypandoc"], dict) and has_pypandoc is None:
        logger.warning(
            "'pypandoc' not available. Using Sphinx-Gallery to "
            "convert rst text blocks to markdown for .ipynb files."
        )
        gallery_conf["pypandoc"] = False
    elif isinstance(gallery_conf["pypandoc"], dict):
        logger.info(
            "Using pandoc version: %s to convert rst text blocks to "
            "markdown for .ipynb files",
            version,
        )
    else:
        logger.info(
            "Using Sphinx-Gallery to convert rst text blocks to "
            "markdown for .ipynb files."
        )
    if isinstance(pypandoc, dict):
        accepted_keys = ("extra_args", "filters")
        for key in pypandoc:
            if key not in accepted_keys:
                raise ConfigError(
                    "'pypandoc' only accepts the following key "
                    f"values: {accepted_keys}, got: {key}."
                )


def _fill_gallery_conf_defaults(sphinx_gallery_conf, app=None, check_keys=True):
    """Handle user configs, update default gallery configs and check values."""
    gallery_conf = copy.deepcopy(DEFAULT_GALLERY_CONF)
    _check_extra_config_keys(gallery_conf, sphinx_gallery_conf, check_keys)

    gallery_conf.update(sphinx_gallery_conf)
    # XXX anything that can only be a bool (rather than str) should probably be
    # evaluated this way as it allows setting via -D on the command line
    for key in (
        "promote_jupyter_magic",
        "run_stale_examples",
    ):
        gallery_conf[key] = _bool_eval(gallery_conf[key])

    gallery_conf["default_role"] = ""
    gallery_conf["source_suffix"] = {".rst": "restructuredtext"}
    if app is not None:
        if app.config["default_role"]:
            gallery_conf["default_role"] = app.config["default_role"]
        gallery_conf["source_suffix"] = app.config["source_suffix"]
        if isinstance(gallery_conf["source_suffix"], str):
            gallery_conf["source_suffix"] = {gallery_conf["source_suffix"]: None}

    # Check capture_repr
    _check_config_type(gallery_conf, "capture_repr", (tuple, list), str_to_list=True)
    supported_reprs = ["__repr__", "__str__", "_repr_html_"]
    for rep in gallery_conf["capture_repr"]:
        if rep not in supported_reprs:
            raise ConfigError(
                "All entries in 'capture_repr' must be one "
                f"of {supported_reprs}, got: {rep}"
            )
    # Check ignore_repr_types
    _check_config_type(gallery_conf, "ignore_repr_types", str)

    # Check parallel
    _check_config_type(gallery_conf, "parallel", (bool, int))
    if gallery_conf["parallel"] is True:
        gallery_conf["parallel"] = app.parallel
    if gallery_conf["parallel"] == 1:
        gallery_conf["parallel"] = False
    if gallery_conf["parallel"]:
        try:
            import joblib  # noqa
        except Exception:
            raise ValueError("joblib must be importable when parallel mode is enabled")

    # deal with show_memory
    _get_call_memory_and_base(gallery_conf, update=True)

    # check callables
    for key in (
        "image_scrapers",
        "reset_argv",
        "minigallery_sort_order",
        "reset_modules",
    ):
        if key == "minigallery_sort_order" and gallery_conf[key] is None:
            continue
        _get_callables(gallery_conf, key)

    # Here we try to set up matplotlib but don't raise an error,
    # we will raise an error later when we actually try to use it
    # (if we do so) in scrapers.py.
    # In principle we could look to see if there is a matplotlib scraper
    # in our scrapers list, but this would be backward incompatible with
    # anyone using or relying on our Agg-setting behavior (e.g., for some
    # custom matplotlib SVG scraper as in our docs).
    # Eventually we can make this a config var like matplotlib_agg or something
    # if people need us not to set it to Agg.
    try:
        _import_matplotlib()
    except (ImportError, ValueError):
        pass

    # Check for srcset hidpi images
    _check_image_srcset(gallery_conf)
    # Check `compress_images`
    _check_compress_images(gallery_conf)
    # Check `matplotlib_animations`
    _check_matplotlib_animations(gallery_conf, app)

    # Check resetters
    _get_callables(gallery_conf, "reset_modules")

    _check_config_type(gallery_conf, "reset_modules_order", str)
    if gallery_conf["reset_modules_order"] not in ["before", "after", "both"]:
        raise ConfigError(
            "reset_modules_order must be in"
            "['before', 'after', 'both'], "
            f"got {gallery_conf['reset_modules_order']!r}"
        )

    # Ensure the first/last cell text is a string if we have it
    cell_config_keys = ("first_notebook_cell", "last_notebook_cell")
    for conf_key in cell_config_keys:
        _check_config_type(gallery_conf, conf_key, str, allow_none=True)

    # Check pypandoc
    _check_pypandoc_config(gallery_conf)

    gallery_conf["titles"] = {}
    # Ensure 'backreferences_dir' is str, pathlib.Path or None
    _check_config_type(
        gallery_conf,
        "backreferences_dir",
        (str, pathlib.Path),
        allow_none=True,
    )
    # if 'backreferences_dir' is pathlib.Path, make str for Python <=3.5
    # compatibility
    backref = gallery_conf["backreferences_dir"]
    if isinstance(backref, pathlib.Path):
        gallery_conf["backreferences_dir"] = str(backref)

    # binder
    gallery_conf["binder"] = check_binder_conf(gallery_conf["binder"])

    # jupyterlite
    gallery_conf["jupyterlite"] = check_jupyterlite_conf(
        gallery_conf["jupyterlite"],
        app,
    )

    # css
    _check_config_type(gallery_conf, "css", (list, tuple), str_to_list=True)
    for css in gallery_conf["css"]:
        if css not in _KNOWN_CSS:
            raise ConfigError(f"Unknown css {css!r}, must be one of {_KNOWN_CSS!r}")
        if app is not None:  # can be None in testing
            app.add_css_file(css + ".css")

    # check API usage
    _check_config_type(gallery_conf, "api_usage_ignore", str)
    if (
        not isinstance(gallery_conf["show_api_usage"], bool)
        and gallery_conf["show_api_usage"] != "unused"
    ):
        raise ConfigError(
            'gallery_conf["show_api_usage"] must be True, False or "unused", '
            f"got {gallery_conf['show_api_usage']}"
        )

    # check `within_subsection_order`
    _get_callables(
        gallery_conf, "within_subsection_order", src_dir=""
    )  # make sure it works

    _update_gallery_conf_exclude_implicit_doc(gallery_conf)

    return gallery_conf


def get_subsections(srcdir, examples_dir, gallery_conf, check_for_header=True):
    """Return the list of subsections of a gallery.

    Parameters
    ----------
    srcdir : str
        absolute path to directory containing conf.py
    examples_dir : str
        path to the examples directory relative to conf.py
    gallery_conf : Dict[str, Any]
        Sphinx-Gallery configuration dictionary.
    check_for_header : bool
        only return subfolders that contain a GALLERY_HEADER file, default True

    Returns
    -------
    out : list
        sorted list of gallery subsection folder names
    """
    if gallery_conf["subsection_order"] is None:
        sortkey = None
    else:
        subsec_order = gallery_conf.get("subsection_order")
        if isinstance(subsec_order, list):
            gallery_conf["subsection_order"] = ExplicitOrder(subsec_order)
        (sortkey,) = _get_callables(gallery_conf, "subsection_order")
    subfolders = [subfolder for subfolder in os.listdir(examples_dir)]
    if check_for_header:
        subfolders = [
            subfolder
            for subfolder in subfolders
            # Return is not `None` only when a gallery head file is found
            if _get_gallery_header(
                os.path.join(examples_dir, subfolder), gallery_conf, raise_error=False
            )
            is not None
        ]
    else:
        # just make sure its a directory, that is not `__pycache__`
        subfolders = [
            subfolder
            for subfolder in subfolders
            if (
                subfolder != "__pycache__"
                and os.path.isdir(os.path.join(examples_dir, subfolder))
            )
        ]

    base_examples_dir_path = os.path.relpath(examples_dir, srcdir)
    subfolders_with_path = [
        os.path.join(base_examples_dir_path, item) for item in subfolders
    ]
    sorted_subfolders = sorted(subfolders_with_path, key=sortkey)

    return [
        subfolders[i]
        for i in [subfolders_with_path.index(item) for item in sorted_subfolders]
    ]


def _prepare_sphx_glr_dirs(gallery_conf, srcdir):
    """Creates necessary folders for sphinx_gallery files."""
    examples_dirs = gallery_conf["examples_dirs"]
    gallery_dirs = gallery_conf["gallery_dirs"]

    if not isinstance(examples_dirs, list):
        examples_dirs = [examples_dirs]

    if not isinstance(gallery_dirs, list):
        gallery_dirs = [gallery_dirs]

    if len(examples_dirs) != len(gallery_dirs):
        logger.warning(
            "'examples_dirs' and 'gallery_dirs' are of different lengths. "
            "Surplus entries will be ignored."
        )

    if bool(gallery_conf["backreferences_dir"]):
        backreferences_dir = os.path.join(srcdir, gallery_conf["backreferences_dir"])
        os.makedirs(backreferences_dir, exist_ok=True)

    return list(zip(examples_dirs, gallery_dirs))


def _filter_tags(subsection_index_content):
    """Filter out tags from `subsection_index_content`."""
    tag_regex = r"^\.\.(\s+)\_(.+)\:(\s*)$"
    subsection_index_content = "\n".join(
        [
            line
            for line in subsection_index_content.splitlines()
            if re.match(tag_regex, line) is None
        ]
        + [""]
    )
    return subsection_index_content


def _finish_index_rst(
    app,
    gallery_conf,
    indexst,
    sg_root_index,
    subsection_index_files,
    gallery_dir_abs_path,
):
    """Add toctree, download and signature, if req, to index and write file."""
    # Generate toctree containing subsection index files
    if (
        sg_root_index
        and gallery_conf["nested_sections"] is True
        and len(subsection_index_files) > 0
    ):
        subsections_toctree = _format_toctree(
            subsection_index_files, includehidden=True
        )
        indexst += subsections_toctree

    # Always generate download zipfiles, only add to index.rst if required
    if gallery_conf["download_all_examples"]:
        download_fhindex = generate_zipfiles(
            gallery_dir_abs_path, app.builder.srcdir, gallery_conf
        )
        if sg_root_index:
            indexst += download_fhindex

    if sg_root_index:
        # Signature
        if app.config.sphinx_gallery_conf["show_signature"]:
            indexst += SPHX_GLR_SIG
        # Write index to file
        index_rst_new = os.path.join(gallery_dir_abs_path, "index.rst.new")
        with codecs.open(index_rst_new, "w", encoding="utf-8") as fhindex:
            fhindex.write(indexst)
        _replace_md5(index_rst_new, mode="t")


def _build_recommender(gallery_conf, gallery_dir_abs_path, subsecs):
    """Build recommender and write recommendations."""
    if gallery_conf["recommender"]["enable"]:
        try:
            import numpy as np  # noqa: F401
        except ImportError:
            raise ConfigError("gallery_conf['recommender'] requires numpy")

        recommender_params = copy.deepcopy(gallery_conf["recommender"])
        recommender_params.pop("enable")
        recommender_params.pop("rubric_header", None)
        recommender = ExampleRecommender(**recommender_params)

        gallery_py_files = []
        # root and subsection directories containing python examples
        gallery_directories = [gallery_dir_abs_path] + subsecs
        for current_dir in gallery_directories:
            src_dir = os.path.join(gallery_dir_abs_path, current_dir)
            # sort python files to have a deterministic input across call
            py_files = sorted(
                # NOTE we don't take account of `ignore_pattern` and ignore
                # ext in `example_extensions`
                [fname for fname in Path(src_dir).iterdir() if fname.suffix == ".py"],
                key=_get_callables(gallery_conf, "within_subsection_order", src_dir)[0],
            )
            gallery_py_files.append(
                [os.path.join(src_dir, fname) for fname in py_files]
            )
        # flatten the list of list
        gallery_py_files = list(chain.from_iterable(gallery_py_files))

        recommender.fit(gallery_py_files)
        for fname in gallery_py_files:
            _write_recommendations(recommender, fname, gallery_conf)


def _log_costs(costs, gallery_conf):
    """Log computation time."""
    logger.info("computation time summary:", color="white")
    lines, lens = _format_for_writing(
        costs, src_dir=gallery_conf["src_dir"], kind="console"
    )
    for name, t, m in lines:
        text = (f"    - {name}:   ").ljust(lens[0] + 10)
        if t is None:
            text += "(not run)"
            logger.info(text)
        else:
            t_float = float(t.split()[0])
            if t_float >= gallery_conf["min_reported_time"]:
                text += t.rjust(lens[1]) + "   " + m.rjust(lens[2])
                logger.info(text)


def generate_gallery_rst(app):
    """Generate the Main examples gallery reStructuredText.

    Fill Sphinx-Gallery configuration and scan example directories
    (up to one level depth of sub-directory) to generate example reST files.

    Iterate through each example directory and any of its sub-directories
    (creates sub-sections) that has a header/index file.
    Generate gallery example ReST files and `index.rst` file(s).

    If `nested_sections=True` we generate `index.rst` files for all
    sub-directories, which includes toctree linking to all sub-dir examples.
    The root example directory `index.rst` file will contain, in sequence,:

    * root gallery header then thumbnails,
    * toctree linking all examples in root gallery,
    * sub-section header followed by sub-section thumbnails, for all subsections,
    * a second final toctree, at the end of the file, linking to all sub-section
      index files.

    If `nested_sections=True` we generate a single `index.rst` file per
    example directory. It will contain headers for the root gallery and
    each sub-section, with each header followed by a toctree linking to
    every example in the root gallery/sub-section.
    """
    gallery_conf = app.config.sphinx_gallery_conf
    extra = ""
    if gallery_conf["parallel"]:
        extra = f" (with parallel={gallery_conf['parallel']})"
    logger.info(f"generating gallery{extra}...", color="white")

    seen_backrefs = set()

    costs = []
    workdirs = _prepare_sphx_glr_dirs(gallery_conf, app.builder.srcdir)

    # Check for duplicate filenames to make sure linking works as expected
    examples_dirs = [ex_dir for ex_dir, _ in workdirs]
    _collect_gallery_files(examples_dirs, gallery_conf, check_filenames=True)

    backrefs_all = {}

    for examples_dir, gallery_dir in workdirs:
        examples_dir_abs_path = os.path.join(app.builder.srcdir, examples_dir)
        gallery_dir_abs_path = os.path.join(app.builder.srcdir, gallery_dir)

        # Create example rst files for root gallery directory examples
        # (excl. sub-dir examples) and fetch gallery header for root index.rst
        (
            _,
            this_content,
            this_costs,
            this_toctree_items,
            backrefs_root,
        ) = generate_dir_rst(
            examples_dir_abs_path,
            gallery_dir_abs_path,
            gallery_conf,
            seen_backrefs,
            is_subsection=False,
        )

        _combine_backreferences(backrefs_all, backrefs_root)

        # `this_content` is None when user provides own index.rst
        sg_root_index = this_content is not None
        costs += this_costs
        write_computation_times(gallery_conf, gallery_dir_abs_path, this_costs)

        # `indexst` variable must exist, as passed to `_finish_index_rst`
        indexst = ""
        # Create root gallery index.rst
        if sg_root_index:
            # :orphan: to suppress "not included in TOCTREE" sphinx warnings
            indexst = ":orphan:\n\n" + this_content
            # Write toctree with gallery items from gallery root folder
            if len(this_toctree_items) > 0:
                this_toctree = _format_toctree(this_toctree_items)
                indexst += this_toctree

        # list all paths to subsection index files in this array
        subsection_index_files = []
        subsecs = get_subsections(
            app.builder.srcdir,
            examples_dir_abs_path,
            gallery_conf,
            check_for_header=sg_root_index,
        )
        for subsection in subsecs:
            src_dir = os.path.join(examples_dir_abs_path, subsection)
            target_dir = os.path.join(gallery_dir_abs_path, subsection)
            subsection_index_files.append(
                "/".join(["", gallery_dir, subsection, "index.rst"]).replace(
                    os.sep, "/"
                )  # fwd slashes needed in rst
            )

            (
                subsection_index_path,
                subsection_index_content,
                subsection_costs,
                subsection_toctree_filenames,
                backrefs_subsec,
            ) = generate_dir_rst(src_dir, target_dir, gallery_conf, seen_backrefs)

            _combine_backreferences(backrefs_all, backrefs_subsec)

            has_subsection_header = False
            if subsection_index_content:
                # Filter tags to prevent tag duplication across the documentation
                indexst += _filter_tags(subsection_index_content)
                has_subsection_header = True

            # Write subsection toctree containing all filenames, if req.
            if (
                sg_root_index
                and not gallery_conf["nested_sections"]
                and len(subsection_toctree_filenames) > 0
            ):
                subsection_index_toctree = _format_toctree(subsection_toctree_filenames)
                indexst += subsection_index_toctree
            # Otherwise, a new subsection index.rst.new file should
            # have been created and it needs to be parsed
            elif has_subsection_header:
                _replace_md5(subsection_index_path, mode="t")

            costs += subsection_costs
            write_computation_times(gallery_conf, target_dir, subsection_costs)

        # Per gallery - items below run once per gallery
        # Finish index.rst and write to file
        _finish_index_rst(
            app,
            gallery_conf,
            indexst,
            sg_root_index,
            subsection_index_files,
            gallery_dir_abs_path,
        )
        # Build recommendation system
        _build_recommender(gallery_conf, gallery_dir_abs_path, subsecs)

    # Per project - items below run once only (for all galleries)
    # Write a single global sg_execution_times
    write_computation_times(gallery_conf, None, costs)

    # Write backreferences_all to file
    if gallery_conf["backreferences_dir"]:
        _write_json(
            Path(
                gallery_conf["src_dir"],
                gallery_conf["backreferences_dir"],
                "backreferences_all",
            ),
            backrefs_all,
        )

    if gallery_conf["show_api_usage"] is not False:
        _init_api_usage(app.builder.srcdir)
    _finalize_backreferences(seen_backrefs, gallery_conf)

    if gallery_conf["plot_gallery"]:
        _log_costs(costs, gallery_conf)
        # Also create a junit.xml file, useful e.g. on CircleCI
        write_junit_xml(gallery_conf, app.builder.outdir, costs)


SPHX_GLR_ORPHAN = """
:orphan:

.. _{0}:

"""

SPHX_GLR_COMP_TIMES = (
    SPHX_GLR_ORPHAN
    + """
Computation times
=================
"""
)


def _sec_to_readable(t):
    """Convert a number of seconds to a more readable representation."""
    # This will only work for < 1 day execution time
    # And we reserve 2 digits for minutes because presumably
    # there aren't many > 99 minute scripts, but occasionally some
    # > 9 minute ones
    t = datetime(1, 1, 1) + timedelta(seconds=t)
    t = "{:02d}:{:02d}.{:03d}".format(
        t.hour * 60 + t.minute, t.second, int(round(t.microsecond / 1000.0))
    )
    return t


def _cost_key(cost):
    """Cost sorting function."""
    # sort by descending computation time, descending memory, alphabetical name
    return (-cost["t"], -cost["mem"], cost["src_file"])


def _format_for_writing(costs, *, src_dir, kind="rst"):
    """Provide formatted computation summary text.

    Parameters
    ----------
    costs: List[Dict]
        List of dicts of computation costs and paths, see gen_rst.py for details.
    src_dir : pathlib.Path
        The Sphinx source directory.
    kind: 'rst', 'rst-full' or 'console', default='rst'
        Format for printing to 'console' or for writing `sg_execution_times.rst' ('rst'
        for single galleries and 'rst-full' for all galleries).

    Returns
    -------
    lines: List[List[str]]
        Formatted computation text for each example, of format:
        [example_file, time_elapsed, memory_used]

    lens: List[int]
        Character length of each string in `lines`.
    """
    lines = list()
    for cost in sorted(costs, key=_cost_key):
        src_file = cost["src_file"]
        rel_path = os.path.relpath(src_file, src_dir)
        if kind in ("rst", "rst-full"):  # like in sg_execution_times
            target_dir_clean = os.path.relpath(cost["target_dir"], src_dir).replace(
                os.sep, "_"
            )
            paren = rel_path if kind == "rst-full" else os.path.basename(src_file)
            name = ":ref:`sphx_glr_{0}_{1}` (``{2}``)".format(
                target_dir_clean, os.path.basename(src_file), paren
            )
            t = _sec_to_readable(cost["t"])
        else:  # like in generate_gallery
            assert kind == "console"
            name = rel_path
            t = f"{cost['t']:0.2f} sec"
        m = f"{cost['mem']:.1f} MB"
        lines.append([name, t, m])
    lens = [max(x) for x in zip(*[[len(item) for item in cost] for cost in lines])]
    return lines, lens


def write_computation_times(gallery_conf, target_dir, costs):
    """Write computation times to `sg_execution_times.rst`.

    Parameters
    ----------
    gallery_conf : Dict[str, Any]
        Sphinx-Gallery configuration dictionary.
    target_dir : str | None
        Path to directory where example python source file are.
    costs: List[Dict]
        List of dicts of computation costs and paths, see gen_rst.py for details.
    """
    if not gallery_conf["write_computation_times"]:
        return
    total_time = sum(cost["t"] for cost in costs)
    if target_dir is None:  # all galleries together
        out_dir = gallery_conf["src_dir"]
        where = "all galleries"
        kind = "rst-full"
        ref_extra = ""
    else:  # a single gallery
        out_dir = target_dir
        where = os.path.relpath(target_dir, gallery_conf["src_dir"])
        kind = "rst"
        ref_extra = f"{where.replace(os.sep, '_')}_"
    new_ref = f"sphx_glr_{ref_extra}sg_execution_times"
    out_file = Path(out_dir) / "sg_execution_times.rst"
    if out_file.is_file() and total_time == 0:  # a re-run
        return
    with out_file.open("w", encoding="utf-8") as fid:
        fid.write(SPHX_GLR_COMP_TIMES.format(new_ref))
        fid.write(
            f"**{_sec_to_readable(total_time)}** total execution time for "
            f"{len(costs)} file{'s' if len(costs) != 1 else ''} **from {where}**:\n\n"
        )
        lines, lens = _format_for_writing(
            costs,
            src_dir=gallery_conf["src_dir"],
            kind=kind,
        )
        del costs
        # https://datatables.net/examples/styling/bootstrap5.html
        fid.write(  # put it in a container to make the scoped style work
            """\
.. container::

  .. raw:: html

    <style scoped>
    <link href="https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/5.3.0/css/bootstrap.min.css" rel="stylesheet" />
    <link href="https://cdn.datatables.net/1.13.6/css/dataTables.bootstrap5.min.css" rel="stylesheet" />
    </style>
    <script src="https://code.jquery.com/jquery-3.7.0.js"></script>
    <script src="https://cdn.datatables.net/1.13.6/js/jquery.dataTables.min.js"></script>
    <script src="https://cdn.datatables.net/1.13.6/js/dataTables.bootstrap5.min.js"></script>
    <script type="text/javascript" class="init">
    $(document).ready( function () {
        $('table.sg-datatable').DataTable({order: [[1, 'desc']]});
    } );
    </script>

  .. list-table::
   :header-rows: 1
   :class: table table-striped sg-datatable

   * - Example
     - Time
     - Mem (MB)
"""  # noqa: E501
        )
        # Need at least one entry or Sphinx complains
        for ex, t, mb in lines or [["N/A", "N/A", "N/A"]]:
            fid.write(
                f"""\
   * - {ex}
     - {t}
     - {mb.rsplit(maxsplit=1)[0]}
"""
            )  # remove the "MB" from the right


def write_api_entries(app, what, name, obj, options, lines):
    """Write api entries to `_sg_api_entries` configuration.

    To connect to `autodoc-process-docstring` event.

    Parameters
    ----------
    app :
        The Sphinx application object.
    what: str
        The type of the object which the docstring belongs to. One of
        "module", "class", "exception", "function", "method", "attribute".
    name :
        The fully qualified name of the object.
    obj :
        The object itself.
    options :
        The options given to the directive: an object with attributes inherited_members,
        undoc_members, show_inheritance and no-index that are true if the flag option of
        same name was given to the auto directive.
    lines :
        The lines of the docstring, see above.
    """
    if app.config.sphinx_gallery_conf["show_api_usage"] is False:
        return
    if "_sg_api_entries" not in app.config.sphinx_gallery_conf:
        app.config.sphinx_gallery_conf["_sg_api_entries"] = dict()
    if what not in app.config.sphinx_gallery_conf["_sg_api_entries"]:
        app.config.sphinx_gallery_conf["_sg_api_entries"][what] = set()
    app.config.sphinx_gallery_conf["_sg_api_entries"][what].add(name)


def _init_api_usage(gallery_dir):
    with codecs.open(
        os.path.join(gallery_dir, "sg_api_usage.rst"), "w", encoding="utf-8"
    ):
        pass


# Colors from https://personal.sron.nl/~pault/data/colourschemes.pdf
# 3 Diverging Colour Schemes, Figure 12, plus alpha=AA
API_COLORS = dict(
    edge="#00000080",  # gray (by alpha)
    okay="#98CAE180",  # blue
    bad_1="#FEDA8B80",  # yellow
    bad_2="#F67E4B80",  # orange
    bad_3="#A5002680",  # red
)


def _make_graph(fname, entries, gallery_conf):
    """Make a graph of unused and used API entries.

    The used API entries themselves are documented in the list, so
    for the graph, we'll focus on the number of unused API entries
    per modules. Modules with lots of unused entries (11+) will be colored
    red, those with less (6+) will be colored orange, those with only a few
    (1-5) will be colored yellow and those with no unused entries will be
    colored blue.

    The API entries that are used are shown with one graph per module.
    That way you can see the examples that each API entry is used in
    for that module (if this was done for the whole project at once,
    the graph would get too large very large quickly).

    Parameters
    ----------
    fname: str
        Path to '*sg_api_unused.dot' file.
    entries: Dict[str, List] or List[str]
        Used (List) or unused (Dict) API entries.
    gallery_conf : Dict[str, Any]
        Sphinx-Gallery configuration dictionary.
    """
    import graphviz

    dg = graphviz.Digraph(
        filename=fname,
        graph_attr={
            "overlap": "scale",
            "pad": "0.5",
        },
        node_attr={
            "color": API_COLORS["okay"],
            "style": "filled",
            "fontsize": "20",
            "shape": "box",
            "fontname": "Open Sans,Arial",
        },
    )

    if isinstance(entries, list):
        connections = set()
        lut = dict()  # look up table for connections so they don't repeat
        structs = [entry.split(".") for entry in entries]
        for struct in sorted(structs, key=len):
            for level in range(len(struct) - 2):
                if (struct[level], struct[level + 1]) in connections:
                    continue
                connections.add((struct[level], struct[level + 1]))
                node_from = (
                    lut[struct[level]] if struct[level] in lut else struct[level]
                )
                dg.node(node_from)
                node_to = struct[level + 1]
                node_kwargs = dict()
                # count, don't show leaves
                if len(struct) - 3 == level:
                    leaf_count = 0
                    for struct2 in structs:
                        # find structures of the same length as struct
                        if len(struct2) != level + 3:
                            continue
                        # find structures with two entries before
                        # the leaf that are the same as struct
                        if all(
                            [
                                struct2[level2] == struct[level2]
                                for level2 in range(level + 2)
                            ]
                        ):
                            leaf_count += 1
                    node_to += f" ({leaf_count})"
                    lut[struct[level + 1]] = node_to
                    if leaf_count > 10:
                        color_key = "bad_3"
                    elif leaf_count > 5:
                        color_key = "bad_2"
                    else:
                        color_key = "bad_1"
                    node_kwargs["color"] = API_COLORS[color_key]
                dg.node(node_to, **node_kwargs)
                dg.edge(node_from, node_to, color=API_COLORS["edge"])
        # add modules with all API entries
        for module in gallery_conf["_sg_api_entries"].get("module", []):
            struct = module.split(".")
            for i in range(len(struct) - 1):
                if struct[i + 1] not in lut:
                    dg.edge(struct[i], struct[i + 1])
    else:
        assert isinstance(entries, dict)
        for entry, refs in entries.items():
            dg.node(entry)
            for ref in refs:
                dg.node(ref, color=API_COLORS["bad_1"])
                dg.edge(entry, ref, color=API_COLORS["edge"])
    dg.save()


def write_api_entry_usage(app, docname, source):
    """Write an html page describing which API entries are used and unused.

    To document and graph only those API entries that are used by
    autodoc, we have to wait for autodoc to finish and hook into the
    ``source-read`` event. This intercepts the text from the rst such
    that it can be modified. Since, we only touched an empty file,
    we have to add 1) a list of all the API entries that are unused
    and a graph of the number of unused API entries per module and 2)
    a list of API entries that are used in examples, each with a sub-list
    of which examples that API entry is used in, and a graph that
    connects all of the API entries in a module to the examples
    that they are used in.

    Parameters
    ----------
    app :
        The Sphinx application object.
    docname :
        Docname of the document currently being parsed.
    source :
        List whose single element is the contents of the source file
    """
    docname = docname or ""  # can be None on Sphinx 7.2
    if docname != "sg_api_usage":
        return
    gallery_conf = app.config.sphinx_gallery_conf
    if gallery_conf["show_api_usage"] is False:
        return
    # since this is done at the gallery directory level (as opposed
    # to in a gallery directory, e.g. auto_examples), it runs last
    # which means that all the api entries will be in gallery_conf

    # Always write at least the title
    source[0] = SPHX_GLR_ORPHAN.format("sphx_glr_sg_api_usage")
    title = "Unused API Entries"
    source[0] += title + "\n" + "^" * len(title) + "\n\n"
    if (
        "_sg_api_entries" not in gallery_conf
        or gallery_conf["backreferences_dir"] is None
    ):
        source[0] += "No API entries found, not computed.\n\n"
        return
    backreferences_dir = os.path.join(
        gallery_conf["src_dir"], gallery_conf["backreferences_dir"]
    )

    example_files = set.union(
        *[
            gallery_conf["_sg_api_entries"][obj_type]
            for obj_type in ("class", "method", "function")
            if obj_type in gallery_conf["_sg_api_entries"]
        ]
    )

    if len(example_files) == 0:
        source[0] += "No examples run, not computed.\n\n"
        return

    def get_entry_type(entry):
        if entry in gallery_conf["_sg_api_entries"].get("class", []):
            return "class"
        elif entry in gallery_conf["_sg_api_entries"].get("method", []):
            return "meth"
        else:
            assert entry in gallery_conf["_sg_api_entries"]["function"]
            return "func"

    # find used and unused API entries
    unused_api_entries = list()
    used_api_entries = dict()
    backreferences_all = _read_json(Path(backreferences_dir, "backreferences_all.json"))
    src_dir = gallery_conf["src_dir"]
    for entry in example_files:
        # don't include built-in methods etc.
        if re.match(gallery_conf["api_usage_ignore"], entry) is not None:
            continue
        # check if backreferences empty
        backref_entry = backreferences_all.get(entry, None)
        if backref_entry is None:
            unused_api_entries.append(entry)
        else:
            used_api_entries[entry] = list()
            for br in backref_entry:
                # br[2] = abs path to target directory
                example_path = Path(br[2], br[0]).relative_to(src_dir)
                ref_name = str(example_path).replace(os.sep, "_")
                used_api_entries[entry].append(f"sphx_glr_{ref_name}")

    for entry in sorted(unused_api_entries):
        source[0] += f"- :{get_entry_type(entry)}:`{entry}`\n"
    source[0] += "\n\n"

    has_graphviz = _has_graphviz()
    if has_graphviz and unused_api_entries:
        source[0] += (
            ".. graphviz:: ./sg_api_unused.dot\n"
            "    :alt: API unused entries graph\n"
            "    :layout: neato\n\n"
        )

    used_count = len(used_api_entries)
    total_count = used_count + len(unused_api_entries)
    used_percentage = used_count / max(total_count, 1)  # avoid div by zero
    source[0] += (
        "\nAPI entries used: "
        f"{round(used_percentage * 100, 2)}% "
        f"({used_count}/{total_count})\n\n"
    )

    if has_graphviz and unused_api_entries:
        _make_graph(
            os.path.join(app.builder.srcdir, "sg_api_unused.dot"),
            unused_api_entries,
            gallery_conf,
        )

    if gallery_conf["show_api_usage"] is True and used_api_entries:
        title = "Used API Entries"
        source[0] += title + "\n" + "^" * len(title) + "\n\n"
        for entry in sorted(used_api_entries):
            source[0] += f"- :{get_entry_type(entry)}:`{entry}`\n\n"
            for ref in used_api_entries[entry]:
                source[0] += f"  - :ref:`{ref}`\n"
            source[0] += "\n\n"

        if has_graphviz:
            used_modules = {entry.split(".")[0] for entry in used_api_entries}
            for module in sorted(used_modules):
                source[0] += (
                    f"{module}\n" + "^" * len(module) + "\n\n"
                    f".. graphviz:: ./{module}_sg_api_used.dot\n"
                    f"    :alt: {module} usage graph\n"
                    "    :layout: neato\n\n"
                )

            for module in used_modules:
                logger.info("Making API usage graph for %s", module)
                # select and format entries for this module
                entries = dict()
                for entry, ref in used_api_entries.items():
                    if entry.split(".")[0] == module:
                        entry = entry.replace("sphx_glr_", "")
                        # remove prefix
                        for target_dir in gallery_conf["gallery_dirs"]:
                            if entry.startswith(target_dir):
                                entry = entry[len(target_dir) + 1 :]
                _make_graph(
                    os.path.join(app.builder.srcdir, f"{module}_sg_api_used.dot"),
                    entries,
                    gallery_conf,
                )


def clean_api_usage_files(app, exception):
    """Remove api usage .dot files.

    To connect to 'build-finished' event.
    """
    if os.path.isfile(os.path.join(app.builder.srcdir, "sg_api_usage.rst")):
        os.remove(os.path.join(app.builder.srcdir, "sg_api_usage.rst"))
    if os.path.isfile(os.path.join(app.builder.srcdir, "sg_api_unused.dot")):
        os.remove(os.path.join(app.builder.srcdir, "sg_api_unused.dot"))
    for file in os.listdir(app.builder.srcdir):
        if "sg_api_used.dot" in file:
            os.remove(os.path.join(app.builder.srcdir, file))


def write_junit_xml(gallery_conf, target_dir, costs):
    """Write JUnit XML file of example run times, successes, and failures.

    Parameters
    ----------
    gallery_conf : Dict[str, Any]
        Sphinx-Gallery configuration dictionary.
    target_dir : Union[str, pathlib.Path]
        Build directory.
    costs: List[Tuple[Tuple[float], str]]
        List of dicts of computation costs and paths, see gen_rst.py for details.
    """
    if not gallery_conf["junit"] or not gallery_conf["plot_gallery"]:
        return
    failing_as_expected, failing_unexpectedly, passing_unexpectedly = _parse_failures(
        gallery_conf
    )
    n_tests = 0
    n_failures = 0
    n_skips = 0
    elapsed = 0.0
    src_dir = gallery_conf["src_dir"]
    output = ""
    for cost in costs:
        t, fname = cost["t"], cost["src_file"]
        if not any(
            fname in x
            for x in (
                gallery_conf["passing_examples"],
                failing_unexpectedly,
                failing_as_expected,
                passing_unexpectedly,
            )
        ):
            continue  # not subselected by our regex
        title = gallery_conf["titles"][fname]
        output += (
            '<testcase classname={!s} file={!s} line="1" name={!s} time="{!r}">'.format(
                quoteattr(os.path.splitext(os.path.basename(fname))[0]),
                quoteattr(os.path.relpath(fname, src_dir)),
                quoteattr(title),
                t,
            )
        )
        if fname in failing_as_expected:
            output += '<skipped message="expected example failure"></skipped>'
            n_skips += 1
        elif fname in failing_unexpectedly or fname in passing_unexpectedly:
            if fname in failing_unexpectedly:
                traceback = gallery_conf["failing_examples"][fname]
            else:  # fname in passing_unexpectedly
                traceback = "Passed even though it was marked to fail"
            n_failures += 1
            output += "<failure message={!s}>{!s}</failure>".format(
                quoteattr(traceback.splitlines()[-1].strip()), escape(traceback)
            )
        output += "</testcase>"
        n_tests += 1
        elapsed += t
    output += "</testsuite>"
    output = (
        '<?xml version="1.0" encoding="utf-8"?>'
        '<testsuite errors="0" failures="{}" name="sphinx-gallery" '
        'skipped="{}" tests="{}" time="{}">'.format(
            n_failures, n_skips, n_tests, elapsed
        )
    ) + output
    # Actually write it
    fname = os.path.normpath(os.path.join(target_dir, gallery_conf["junit"]))
    junit_dir = os.path.dirname(fname)
    os.makedirs(junit_dir, exist_ok=True)
    with codecs.open(fname, "w", encoding="utf-8") as fid:
        fid.write(output)


def touch_empty_backreferences(app, what, name, obj, options, lines):
    """Generate empty back-reference example files.

    This avoids inclusion errors/warnings if there are no gallery
    examples for a class / module that is being parsed by autodoc.
    """
    if not bool(app.config.sphinx_gallery_conf["backreferences_dir"]):
        return

    examples_path = os.path.join(
        app.srcdir,
        app.config.sphinx_gallery_conf["backreferences_dir"],
        f"{name}.examples",
    )

    if not os.path.exists(examples_path):
        # touch file
        open(examples_path, "w").close()


def _expected_failing_examples(gallery_conf):
    return {
        os.path.normpath(os.path.join(gallery_conf["src_dir"], path))
        for path in gallery_conf["expected_failing_examples"]
    }


def _parse_failures(gallery_conf):
    """Split the failures."""
    failing_examples = set(gallery_conf["failing_examples"])
    expected_failing_examples = _expected_failing_examples(gallery_conf)
    failing_as_expected = failing_examples.intersection(expected_failing_examples)
    failing_unexpectedly = failing_examples.difference(expected_failing_examples)
    passing_unexpectedly = expected_failing_examples.difference(failing_examples)
    # filter from examples actually run
    passing_unexpectedly = set(
        src_file
        for src_file in passing_unexpectedly
        if re.search(gallery_conf["filename_pattern"], src_file)
    )
    return failing_as_expected, failing_unexpectedly, passing_unexpectedly


def summarize_failing_examples(app, exception):
    """Collects the list of falling examples and prints them with a traceback.

    Raises ValueError if there where failing examples.
    """
    if exception is not None:
        return

    # Under no-plot Examples are not run so nothing to summarize
    if not app.config.sphinx_gallery_conf["plot_gallery"]:
        logger.info(
            'Sphinx-Gallery gallery_conf["plot_gallery"] was '
            "False, so no examples were executed.",
            color="brown",
        )
        return

    gallery_conf = app.config.sphinx_gallery_conf
    failing_as_expected, failing_unexpectedly, passing_unexpectedly = _parse_failures(
        gallery_conf
    )

    idt = "    "
    if failing_as_expected:
        logger.info(
            bold(blue(f"Examples failing as expected ({len(failing_as_expected)}):"))
        )
        for fail_example in failing_as_expected:
            path = os.path.relpath(fail_example, gallery_conf["src_dir"])
            logger.info(
                f"{bold(blue(path))} failed leaving traceback:\n\n"
                f"{indent(gallery_conf['failing_examples'][fail_example], idt)}"
            )

    fail_msgs = []
    if failing_unexpectedly:
        fail_msgs.append(
            bold(red(f"Unexpected failing examples ({len(failing_unexpectedly)}):\n"))
        )
        for fail_example in failing_unexpectedly:
            path = os.path.relpath(fail_example, gallery_conf["src_dir"])
            fail_msgs.append(
                f"    {bold(red(path))} failed leaving traceback:\n\n"
                f"{indent(gallery_conf['failing_examples'][fail_example], idt)}"
            )

    if passing_unexpectedly:
        paths = [
            os.path.relpath(p, gallery_conf["src_dir"]) for p in passing_unexpectedly
        ]
        fail_msgs.append(
            bold(red(f"Examples expected to fail, but not failing ({len(paths)}):\n\n"))
            + red("\n".join(indent(p, idt) for p in paths))
            + "\n\nPlease remove these examples from "
            + "sphinx_gallery_conf['expected_failing_examples'] "
            + "in your conf.py file."
        )

    # standard message
    n_good = len(gallery_conf["passing_examples"])
    n_tot = len(gallery_conf["failing_examples"]) + n_good
    n_stale = len(gallery_conf["stale_examples"])
    logger.info(
        "\nSphinx-Gallery successfully executed %d out of %d "
        "file%s subselected by:\n\n"
        '    gallery_conf["filename_pattern"] = %r\n'
        '    gallery_conf["ignore_pattern"]   = %r\n'
        "\nafter excluding %d file%s that had previously been run "
        "(based on MD5).\n",
        n_good,
        n_tot,
        "s" if n_tot != 1 else "",
        gallery_conf["filename_pattern"],
        gallery_conf["ignore_pattern"],
        n_stale,
        "s" if n_stale != 1 else "",
        color="brown",
    )

    if fail_msgs:
        fail_message = bold(
            purple(
                "Here is a summary of the problems encountered "
                "when running the examples:\n\n"
                + "\n".join(fail_msgs)
                + "\n"
                + "-" * 79
            )
        )
        if gallery_conf["only_warn_on_example_error"]:
            logger.warning(fail_message)
        else:
            raise ExtensionError(fail_message)


def get_default_config_value(key):
    """Get default configuration function."""

    def default_getter(conf):
        return conf["sphinx_gallery_conf"].get(key, DEFAULT_GALLERY_CONF[key])

    return default_getter


def fill_gallery_conf_defaults(app, config, check_keys=True):
    """Check the sphinx-gallery config and set its defaults.

    This is called early at config-inited, so that all the rest of the code can
    do things like ``sphinx_gallery_conf['binder']['use_jupyter_lab']``, even
    if the keys have not been set explicitly in conf.py.
    """
    new_sphinx_gallery_conf = _fill_gallery_conf_defaults(
        config.sphinx_gallery_conf, app=app, check_keys=check_keys
    )
    config.sphinx_gallery_conf = new_sphinx_gallery_conf
    config.html_static_path.append(glr_path_static())

    # This must be done in "config-inited" for Sphinx 5 and 6 because otherwise the
    # `templates_path` will be overwritten and Sphinx will not be able to discover our
    # component templates
    config.templates_path.append(str(Path(__file__).parent / "components"))


def update_gallery_conf_builder_inited(app):
    """Update the the sphinx-gallery config at builder-inited."""
    plot_gallery = _bool_eval(app.builder.config.plot_gallery)
    src_dir = app.builder.srcdir
    abort_on_example_error = _bool_eval(app.builder.config.abort_on_example_error)
    _update_gallery_conf_builder_inited(
        app.config.sphinx_gallery_conf,
        src_dir,
        plot_gallery=plot_gallery,
        abort_on_example_error=abort_on_example_error,
        builder_name=app.builder.name,
    )


def setup_template_link_getters(app, pagename, templatename, context, doctree):
    """Set up the getters for download and launcher links.

    The getters are added to the sphinx context so as to be used in templates.
    """

    def _find_containers_with_class(class_name):
        if doctree is None:
            return iter([])

        return doctree.findall(
            lambda x: isinstance(x, nodes.container)
            and class_name in x.attributes.get("classes", [])
        )

    def get_download_links():
        """Get the download links for the example.

        This function relies on `_find_containers_with_class` which in turn relies on
        `doctree` provided in the `html-page-context` event. This function will then
        be added to the context of the page and can be accessed in the template.

        This returns a dictionary with keys in ["python", "jupyter", "zip"], depending
        on their availability. The values contain:
        - link: The relative path to the download file
        - label: The "Download {label}" text
        - title: The title to show when hovering over the link
        """
        links = {}
        for key, label in [
            ("python", "source code"),
            ("jupyter", "Jupyter notebook"),
            ("zip", "zipped"),
        ]:
            containers = _find_containers_with_class(f"sphx-glr-download-{key}")
            if container := next(containers, None):
                attrs = container.children[0].children[0].attributes
                if link := attrs.get("filename"):
                    links[key] = {
                        "link": f"_downloads/{link}",
                        "label": label,
                        "title": attrs.get("reftarget"),
                    }
        return links

    def get_launcher_links():
        """Get the launcher links for the example.

        This function relies on `_find_containers_with_class` which in turn relies on
        `doctree` provided in the `html-page-context` event. This function will then
        be added to the context of the page and can be accessed in the template.

        This returns a dictionary with keys in ["lite", "binder"], depending on their
        availability. The values contain:
        - link: The URL to Binder or JupyterLite link of the example
        - img_alt: The alt text for the link badge
        - img_src: The source URL of the link badge
        """
        links = {}
        for key in ["lite", "binder"]:
            containers = _find_containers_with_class(f"{key}-badge")
            if container := next(containers, None):
                anchor = container.children[0]
                image = anchor.children[0]
                if link := anchor.attributes.get("refuri"):
                    links[key] = {
                        "link": link,
                        "img_alt": image.attributes.get("alt"),
                        "img_src": image.attributes.get("uri"),
                    }
        return links

    context["get_download_links"] = get_download_links
    context["get_launcher_links"] = get_launcher_links


def setup(app):
    """Setup Sphinx-Gallery sphinx extension."""
    app.add_config_value("sphinx_gallery_conf", DEFAULT_GALLERY_CONF, "html")
    for key in ["plot_gallery", "abort_on_example_error"]:
        app.add_config_value(key, get_default_config_value(key), "html")

    # Early filling of sphinx_gallery_conf defaults at config-inited
    app.connect("config-inited", fill_gallery_conf_defaults, priority=10)
    # set small priority value, so that pre_configure_jupyterlite_sphinx is
    # called before jupyterlite_sphinx config-inited
    app.connect("config-inited", pre_configure_jupyterlite_sphinx, priority=100)
    # set high priority value, so that post_configure_jupyterlite_sphinx is
    # called after jupyterlite_sphinx config-inited
    app.connect("config-inited", post_configure_jupyterlite_sphinx, priority=900)

    if "sphinx.ext.autodoc" in app.extensions:
        app.connect("autodoc-process-docstring", touch_empty_backreferences)
        app.connect("autodoc-process-docstring", write_api_entries)
        app.connect("source-read", write_api_entry_usage)

    # Add the custom directive
    app.add_directive("minigallery", MiniGallery)
    app.add_directive("image-sg", ImageSg)

    imagesg_addnode(app)

    # Early update of sphinx_gallery_conf at builder-inited
    app.connect("builder-inited", update_gallery_conf_builder_inited, priority=10)
    app.connect("builder-inited", generate_gallery_rst)
    app.connect("build-finished", copy_binder_files)
    app.connect("build-finished", create_jupyterlite_contents)

    app.connect("build-finished", summarize_failing_examples)
    app.connect("build-finished", embed_code_links)
    app.connect("build-finished", clean_api_usage_files)

    app.connect("html-page-context", setup_template_link_getters)

    metadata = {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
        "version": _sg_version,
    }
    return metadata


def setup_module():
    """Hack to stop nosetests running setup() above."""
    pass



"""Pytest fixtures."""

import shutil
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from unittest.mock import Mock

import pytest
import sphinx
from sphinx.application import Sphinx
from sphinx.errors import ExtensionError
from sphinx.util.docutils import docutils_namespace

from sphinx_gallery import docs_resolv, gen_gallery, gen_rst, py_source_parser
from sphinx_gallery.scrapers import _import_matplotlib
from sphinx_gallery.utils import _get_image

INDEX_RST = """
=============
Own index.rst
=============

Own index.rst file.

.. toctree::

    plot_1
    plot_2
    plot_3
"""

NESTED_PY = """\"\"\"
Header
======

Text.
\"\"\"

a = 1
"""

GALLERY_HEADER = """
Gallery header
==============

Some text.
"""


def pytest_report_header(config, startdir=None):
    """Add information to the pytest run header."""
    return f"Sphinx:  {sphinx.__version__} ({sphinx.__file__})"


@pytest.fixture
def gallery_conf(tmpdir):
    """Set up a test sphinx-gallery configuration."""
    app = Mock(
        spec=Sphinx,
        config=dict(source_suffix={".rst": None}, default_role=None),
        extensions=[],
    )
    gallery_conf = gen_gallery._fill_gallery_conf_defaults({}, app=app)
    gen_gallery._update_gallery_conf_builder_inited(gallery_conf, str(tmpdir))
    gallery_conf.update(examples_dir=str(tmpdir), gallery_dir=str(tmpdir))
    return gallery_conf


@pytest.fixture
def log_collector(monkeypatch):
    app = Mock(spec=Sphinx, name="FakeSphinxApp")()
    monkeypatch.setattr(docs_resolv, "logger", app)
    monkeypatch.setattr(gen_gallery, "logger", app)
    monkeypatch.setattr(py_source_parser, "logger", app)
    monkeypatch.setattr(gen_rst, "logger", app)
    yield app


@pytest.fixture
def unicode_sample(tmpdir):
    """Return temporary python source file with Unicode in various places."""
    code_str = b"""# -*- coding: utf-8 -*-
'''
\xc3\x9anicode in header
=================

U\xc3\xb1icode in description
'''

# Code source: \xc3\x93scar N\xc3\xa1jera
# License: BSD 3 clause

import os
path = os.path.join('a','b')

a = 'hei\xc3\x9f'  # Unicode string

import sphinx_gallery.back_references as br
br.identify_names

from sphinx_gallery.back_references import identify_names
identify_names

from sphinx_gallery._dummy import DummyClass
DummyClass().prop

from sphinx_gallery._dummy.nested import NestedDummyClass
NestedDummyClass().prop

import matplotlib.pyplot as plt
_ = plt.figure()

"""

    fname = tmpdir.join("unicode_sample.py")
    fname.write(code_str, "wb")
    return fname.strpath


@pytest.fixture
def req_mpl_jpg(tmpdir, req_mpl, scope="session"):
    """Raise SkipTest if JPEG support is not available."""
    # mostly this is needed because of
    # https://github.com/matplotlib/matplotlib/issues/16083
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    ax.plot(range(10))
    try:
        plt.savefig(str(tmpdir.join("testplot.jpg")))
    except Exception as exp:
        pytest.skip(f"Matplotlib jpeg saving failed: {exp}")
    finally:
        plt.close(fig)


@pytest.fixture(scope="session")
def req_mpl():
    try:
        _import_matplotlib()
    except (ImportError, ValueError):
        pytest.skip("Test requires matplotlib")


@pytest.fixture(scope="session")
def req_pil():
    try:
        _get_image()
    except ExtensionError:
        pytest.skip("Test requires pillow")


@pytest.fixture
def conf_file(request):
    try:
        env = request.node.get_closest_marker("add_conf")
    except AttributeError:  # old pytest
        env = request.node.get_marker("add_conf")
    kwargs = env.kwargs if env else {}
    result = {
        "content": "",
        "extensions": ["sphinx_gallery.gen_gallery"],
    }
    result.update(kwargs)

    return result


@pytest.fixture
def rst_file(request):
    try:
        env = request.node.get_closest_marker("add_rst")
    except AttributeError:  # old pytest
        env = request.node.get_marker("add_rst")
    file = env.kwargs["file"] if env else ""
    return file


class SphinxAppWrapper:
    """Wrapper for sphinx.application.Application.

    This allows control over when the sphinx application is initialized, since
    part of the sphinx-gallery build is done in
    sphinx.application.Application.__init__ and the remainder is done in
    sphinx.application.Application.build.
    """

    def __init__(self, srcdir, confdir, outdir, doctreedir, buildername, **kwargs):
        self.srcdir = srcdir
        self.confdir = confdir
        self.outdir = outdir
        self.doctreedir = doctreedir
        self.buildername = buildername
        self.kwargs = kwargs

    def create_sphinx_app(self):
        """Create Sphinx app."""
        # Avoid warnings about re-registration, see:
        # https://github.com/sphinx-doc/sphinx/issues/5038
        with self.create_sphinx_app_context() as app:
            pass
        return app

    @contextmanager
    def create_sphinx_app_context(self):
        """Create Sphinx app inside context."""
        with docutils_namespace():
            app = Sphinx(
                self.srcdir,
                self.confdir,
                self.outdir,
                self.doctreedir,
                self.buildername,
                **self.kwargs,
            )
            yield app

    def build_sphinx_app(self, *args, **kwargs):
        """Build Sphinx app."""
        with self.create_sphinx_app_context() as app:
            # building should be done in the same docutils_namespace context
            app.build(*args, **kwargs)
        return app


@pytest.fixture
def sphinx_app_wrapper(tmpdir, conf_file, rst_file, req_mpl, req_pil):
    _fixturedir = Path(__file__).parent / "testconfs"
    srcdir = Path(tmpdir) / "config_test"
    shutil.copytree(_fixturedir, srcdir)
    # Copy files to 'examples/' as well because default `examples_dirs` is
    # '../examples' - for tests where we don't update config
    shutil.copytree((_fixturedir / "src"), (Path(tmpdir) / "examples"))
    if rst_file == "own index.rst":
        with open((srcdir / "src" / "index.rst"), "w") as file:
            file.write(INDEX_RST)
    elif rst_file:
        with open((srcdir / "minigallery_test.rst"), "w") as file:
            file.write(rst_file)
        # Add nested gallery
        if "sub_folder/sub_sub_folder" in rst_file:
            dir_path = srcdir / "src" / "sub_folder" / "sub_sub_folder"
            dir_path.mkdir(parents=True)
            with open((dir_path / "plot_nested.py"), "w") as file:
                file.write(NESTED_PY)
            with open((dir_path / "GALLERY_HEADER.rst"), "w") as file:
                file.write(GALLERY_HEADER)

    base_config = f"""
import os
import sphinx_gallery
extensions = {conf_file["extensions"]!r}
exclude_patterns = ['_build', 'src']
source_suffix = '.rst'
master_doc = 'index'
# General information about the project.
project = 'Sphinx-Gallery <Tests>'\n\n
"""
    with open((srcdir / "conf.py"), "w") as conffile:
        conffile.write(base_config + conf_file["content"])

    return SphinxAppWrapper(
        srcdir,
        srcdir,
        (srcdir / "_build"),
        (srcdir / "_build" / "toctree"),
        "html",
        warning=StringIO(),
        status=StringIO(),
    )



"""
Testing backreferences
----------------------

This example runs after plot_future_statements.py (alphabetical ordering within
subsection) and should be unaffected by the ``__future__`` import in
plot_future_statements.py. We should eventually update this script to actually
test this... we require Python 3 nowadays so the ``__future__`` statements there
don't do anything. So for now let's repurpose this to look at some
backreferences. We should probably also change the filename in another PR!
"""

# sphinx_gallery_thumbnail_path = '_static_nonstandard/demo.png'

from sphinx_gallery.scrapers import clean_modules, figure_rst
from sphinx_gallery.sorting import ExplicitOrder

ExplicitOrder([])  # must actually be used to become a backref target!

assert 3 / 2 == 1.5
assert figure_rst([], "") == ""
assert clean_modules(dict(reset_modules=[]), "", "before") is None



"""
.. _ray_trace_example:

Ray Tracing
~~~~~~~~~~~

Single line segment ray tracing for PolyData objects.
"""

import pyvista as pv

# Create source to ray trace
sphere = pv.Sphere(radius=0.85)

# Define line segment
start = [0, 0, 0]
stop = [0.25, 1, 0.5]

# Perform ray trace
points, ind = sphere.ray_trace(start, stop)

# Create geometry to represent ray trace
ray = pv.Line(start, stop)
intersection = pv.PolyData(points)

# Render the result
p = pv.Plotter()
p.add_mesh(
    sphere, show_edges=True, opacity=0.5, color="w", lighting=False, label="Test Mesh"
)
p.add_mesh(ray, color="blue", line_width=5, label="Ray Segment")
p.add_mesh(intersection, color="maroon", point_size=25, label="Intersection Points")
p.add_legend()
p.show()



# Author: Óscar Nájera
# License: 3-clause BSD
r"""Test Sphinx-Gallery gallery generation."""

import inspect
import json
import os
import re
from pathlib import Path

import pytest
from sphinx.application import Sphinx
from sphinx.config import is_serializable
from sphinx.errors import ConfigError, ExtensionError, SphinxWarning

from sphinx_gallery.gen_gallery import (
    _fill_gallery_conf_defaults,
    fill_gallery_conf_defaults,
    write_api_entry_usage,
    write_computation_times,
)
from sphinx_gallery.interactive_example import create_jupyterlite_contents
from sphinx_gallery.utils import (
    _collect_gallery_files,
    _escape_ansi,
    check_duplicate_filenames,
    check_spaces_in_filenames,
)

MINIMAL_HEADER = """
'''
Title
-----
Description.
'''

"""


def test_bad_config():
    """Test that bad config values are caught."""
    sphinx_gallery_conf = dict(example_dir="")
    with pytest.raises(
        ConfigError, match="example_dir.*did you mean 'examples_dirs'?.*"
    ):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)
    sphinx_gallery_conf = dict(n_subsection_order="")
    with pytest.raises(
        ConfigError, match=r"did you mean one of \['subsection_order', 'within_.*"
    ):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)
    sphinx_gallery_conf = dict(within_subsection_order="sphinx_gallery.a.b.Key")
    with pytest.raises(ConfigError, match=r"Unknown string option.*when importing"):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)
    sphinx_gallery_conf = dict(within_subsection_order=1.0)
    with pytest.raises(ConfigError, match="must be callable"):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)
    sphinx_gallery_conf = dict(minigallery_sort_order=int)
    with pytest.raises(ConfigError, match="Got class rather than callable instance"):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)


def test_default_config(sphinx_app_wrapper):
    """Test default Sphinx-Gallery config loaded when extension added to Sphinx."""
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    cfg = sphinx_app.config
    assert cfg.project == "Sphinx-Gallery <Tests>"
    # no duplicate values allowed The config is present already
    with pytest.raises(ExtensionError) as excinfo:
        sphinx_app.add_config_value("sphinx_gallery_conf", "x", True)
    assert "already present" in str(excinfo.value)


def test_serializable(sphinx_app_wrapper):
    """Test that the default config is serializable."""
    bad = list()
    for key, val in _fill_gallery_conf_defaults({}).items():
        if not is_serializable(val):
            bad.append(f"{repr(key)}: {repr(val)}")

    bad = "\n".join(bad)
    assert not bad, f"Non-serializable values found:\n{bad}"


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
def test_no_warning_simple_config(sphinx_app_wrapper):
    """Testing that no warning is issued with a simple config.

    The simple config only specifies input (examples_dirs) and output
    (gallery_dirs) directories.
    """
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    cfg = sphinx_app.config
    assert cfg.project == "Sphinx-Gallery <Tests>"
    build_warn = sphinx_app._warning.getvalue()
    assert build_warn == ""


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': ['src', 'excess'],
    'gallery_dirs': 'ex',
}"""
)
def test_unequal_examples_gallery_dirs(sphinx_app_wrapper):
    """Check warning when 'examples_dirs' and 'gallery_dirs' unequal lengths."""
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    build_warn = sphinx_app._warning.getvalue()
    assert "'examples_dirs' and 'gallery_dirs' are of different lengths" in build_warn


# This changed from passing the ValueError directly to
# raising "sphinx.errors.ConfigError" with "threw an exception"
@pytest.mark.parametrize(
    "err_class, err_match",
    [
        pytest.param(
            ConfigError,
            "Unknown string option for reset_modules",
            id="Resetter unknown",
            marks=pytest.mark.add_conf(
                content="sphinx_gallery_conf={'reset_modules': ('f',)}"
            ),
        ),
        pytest.param(
            ConfigError,
            "reset_modules.* must be callable",
            id="Resetter not callable",
            marks=pytest.mark.add_conf(
                content="sphinx_gallery_conf={'reset_modules': (1.,),}"
            ),
        ),
    ],
)
def test_bad_reset(sphinx_app_wrapper, err_class, err_match):
    with pytest.raises(err_class, match=err_match):
        sphinx_app_wrapper.create_sphinx_app()


@pytest.mark.parametrize(
    "err_class, err_match",
    [
        pytest.param(
            ConfigError,
            "'reset_modules_order' config allowed types",
            id="Resetter unknown",
            marks=pytest.mark.add_conf(
                content=("sphinx_gallery_conf={'reset_modules_order': 1,}")
            ),
        ),
        pytest.param(
            ConfigError,
            "reset_modules_order must be in",
            id="reset_modules_order not valid",
            marks=pytest.mark.add_conf(
                content=("sphinx_gallery_conf={'reset_modules_order': 'invalid',}")
            ),
        ),
    ],
)
def test_bad_reset_modules_order(sphinx_app_wrapper, err_class, err_match):
    with pytest.raises(err_class, match=err_match):
        sphinx_app_wrapper.create_sphinx_app()


@pytest.mark.parametrize(
    "err_class, err_match",
    [
        pytest.param(
            ConfigError,
            "Unknown css",
            id="CSS str error",
            marks=pytest.mark.add_conf(content="sphinx_gallery_conf={'css': ('foo',)}"),
        ),
        pytest.param(
            ConfigError,
            "config allowed types:",
            id="CSS type error",
            marks=pytest.mark.add_conf(content="sphinx_gallery_conf={'css': 1.}"),
        ),
    ],
)
def test_bad_css(sphinx_app_wrapper, err_class, err_match):
    """Test 'css' configuration validation is correct."""
    with pytest.raises(err_class, match=err_match):
        sphinx_app_wrapper.create_sphinx_app()


def test_bad_api():
    """Test that we raise an error for bad API usage arguments."""
    sphinx_gallery_conf = dict(api_usage_ignore=("foo",))
    with pytest.raises(ConfigError, match="'api_usage_ignore' config allowed"):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)
    sphinx_gallery_conf = dict(show_api_usage="foo")
    with pytest.raises(ConfigError, match='.*must be True, False or "unused".*'):
        _fill_gallery_conf_defaults(sphinx_gallery_conf)


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'backreferences_dir': os.path.join('gen_modules', 'backreferences'),
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
def test_config_backreferences(sphinx_app_wrapper):
    """Test no warning is issued under the new configuration."""
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    cfg = sphinx_app.config
    assert cfg.project == "Sphinx-Gallery <Tests>"
    assert cfg.sphinx_gallery_conf["backreferences_dir"] == os.path.join(
        "gen_modules", "backreferences"
    )
    build_warn = sphinx_app._warning.getvalue()
    assert build_warn == ""


def test_duplicate_files_warn(sphinx_app_wrapper):
    """Test for a warning when two files with the same filename exist."""
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()

    files = ["./a/file1.py", "./a/file2.py", "a/file3.py", "./b/file1.py"]
    msg = (
        "Duplicate example file name(s) found. Having duplicate file names "
        "will break some links. List of files: {}"
    )
    m = "['./b/file1.py']"

    # No warning because no overlapping names
    check_duplicate_filenames(files[:-1])
    build_warn = sphinx_app._warning.getvalue()
    assert build_warn == ""

    # Warning because last file is named the same
    check_duplicate_filenames(files)
    build_warn = sphinx_app._warning.getvalue()
    assert msg.format(m) in build_warn


def test_spaces_in_files_warn(sphinx_app_wrapper):
    """Test for a exception when an example filename has a space in it."""
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()

    files = ["./a/file1.py", "./a/file2.py", "./a/file 3.py"]
    msg = (
        "Example file name(s) with spaces found. Having spaces in "
        "file names will break some links. "
        "List of files: {}"
    )
    m = "['./a/file 3.py']"

    # No warning because no filename with space
    check_spaces_in_filenames(files[:-1])
    build_warn = sphinx_app._warning.getvalue()
    assert build_warn == ""

    # Warning because last file has space
    check_spaces_in_filenames(files)
    build_warn = sphinx_app._warning.getvalue()
    assert msg.format(m) in build_warn


def _check_order(sphinx_app, key, expected_order=None):
    """Iterates through sphx-glr-thumbcontainer divs and reads key from the tooltip.

    Used to test that these keys (in index.rst) appear in a specific order. The `.py`
    files include a source-of-truth line indicating the position in the sort order that
    the thumbnail for that file should occur, for the various built-in sorters.

    The regex below extracts info from that source-of-truth line to check that the
    thumbnail ordering was correct --- unless `expected_order` is provided, in which
    case an alternative regex extracts the digit-part of the filename from the crossref,
    and compares the order of crossrefs to `expected_order`.
    """
    index_fname = Path(sphinx_app.outdir, "..", "ex", "index.rst")
    order = list()
    if key is None:
        regex = r".*:ref:`sphx_glr_ex_(?:(?:first|second)-subsection_)?plot_(\d)\.py`"
        locator = ":ref:"
    else:
        regex = rf".*:{key}=(\d):.*"
        locator = "sphx-glr-thumbcontainer"
    with open(index_fname, "r", encoding="utf-8") as fid:
        for line in fid:
            if locator in line:
                order.append(re.match(regex, line).group(1))
    expected_order = expected_order or list("123456789")
    assert order == expected_order


# ======================================================================================
# SUBSECTION AND WITHIN-SUBSECTION SORTING TESTS
# ======================================================================================
# pre-construct the pytest params for the builtin sorters, as they're all very similar
_template_conf_for_builtin_sorters = """
sphinx_gallery_conf = {{
    "examples_dirs": "src",
    "gallery_dirs": "ex",
    "within_subsection_order": "{}",
}}"""
_builtin_sorters = dict(
    lines="NumberOfCodeLinesSortKey",  # the default
    filesize="FileSizeSortKey",
    filename="FileNameSortKey",
    title="ExampleTitleSortKey",
)
_params_for_testing_builtin_sorters = (
    pytest.param(
        kind,  # `sort_key` param
        None,  # `expected_order` (order is extracted from comments in test site pages)
        id=f"within_subsection_sort_by_{kind}",
        marks=pytest.mark.add_conf(
            content=_template_conf_for_builtin_sorters.format(sorter)
        ),
    )
    for kind, sorter in _builtin_sorters.items()
)
# now a more flexible config template, for changing subsection and/or within-subsection
# order (possibly at the same time):
_template_conf = """
{imports}
sphinx_gallery_conf = {{
    "examples_dirs": "src",
    "gallery_dirs": "ex",
    "subsection_order": {subsection_order},
    "within_subsection_order": {within_subsection_order},
}}"""
# fill in the template config with different test cases and wrap in an `add_conf` mark
_subsection_explicit_order = pytest.mark.add_conf(
    content=_template_conf.format(
        imports="from sphinx_gallery.sorting import ExplicitOrder",
        subsection_order='ExplicitOrder(["src/second-subsection", "src/first-subsection"])',
        within_subsection_order='"NumberOfCodeLinesSortKey"',  # this is the default
    )
)
_subsection_explicit_order_list = pytest.mark.add_conf(
    content=_template_conf.format(
        imports="",
        subsection_order='["src/second-subsection", "src/first-subsection"]',
        within_subsection_order="sphinx_gallery.sorting.NumberOfCodeLinesSortKey",
    )
)
_both_custom_fqn = pytest.mark.add_conf(
    content=_template_conf.format(
        imports="",
        subsection_order='"sphinx_gallery.utils._custom_subsection_sorter"',
        within_subsection_order='"sphinx_gallery.utils._custom_example_sorter"',
    )
)
_within_subsection_custom_fqn = pytest.mark.add_conf(
    content=_template_conf.format(
        imports="",
        subsection_order="None",  # the default, AKA, sort by foldername
        within_subsection_order='"sphinx_gallery.utils._custom_example_sorter"',
    )
)
_custom_func = """
from sphinx_gallery.sorting import FunctionSortKey

def custom_sorter(filename):
    ORDER = [
        "plot_3.py",
        "plot_2.py",
        "plot_1.py",
        "plot_6.py",
        "plot_5.py",
        "plot_4.py",
        "plot_9.py",
        "plot_7.py",
        "plot_8.py",
    ]
    return ORDER.index(filename)
"""
_within_subsection_custom_func = pytest.mark.add_conf(
    content=_template_conf.format(
        imports=_custom_func,
        subsection_order="None",  # the default, AKA, sort by foldername
        within_subsection_order="FunctionSortKey(custom_sorter)",
    )
)
_subsection_fqn_within_subsection_custom_func = pytest.mark.add_conf(
    content=_template_conf.format(
        imports=_custom_func,
        subsection_order='"sphinx_gallery.utils._custom_subsection_sorter"',
        within_subsection_order="FunctionSortKey(custom_sorter)",
    )
)


@pytest.mark.parametrize(
    "sort_key,expected_order",
    (
        pytest.param(
            "lines",
            # ↓ natural order except subsection 2 (789) precedes subsection 1 (456)
            list("123789456"),
            id="subsection_sort_by_ExplicitOrder",
            marks=_subsection_explicit_order,
        ),
        pytest.param(
            "lines",
            # ↓ natural order except subsection 2 (789) precedes subsection 1 (456)
            list("123789456"),
            id="subsection_sort_by_list",
            marks=_subsection_explicit_order_list,
        ),
        pytest.param(
            None,
            # ↓ order determined by `utils._CUSTOM_EXAMPLE_ORDER`
            list("132564879"),
            id="within_subsection_sort_by_custom_FQN",
            marks=_within_subsection_custom_fqn,
        ),
        pytest.param(
            None,
            # ↓ order set by `utils._CUSTOM_EXAMPLE_ORDER`, with middle (564) and
            # ↓ last (879) thirds swapped due to subsection reordering
            list("132879564"),
            id="subsection_and_within_subsection_both_sort_by_custom_FQN",
            marks=_both_custom_fqn,
        ),
        pytest.param(
            None,
            # ↓ order set by local `_custom_func` variable within this test file
            list("321654978"),
            id="within_subsection_sort_by_custom_func",
            marks=_within_subsection_custom_func,
        ),
        pytest.param(
            None,
            # ↓ order set by `_custom_func` in this test file, with middle (654) and
            # ↓ last (978) thirds swapped due to subsection reordering
            list("321978654"),
            id="subsection_sort_by_FQN_and_within_subsection_sort_by_custom_func",
            marks=_subsection_fqn_within_subsection_custom_func,
        ),
        *_params_for_testing_builtin_sorters,
    ),
)
def test_example_sorting(sphinx_app_wrapper, sort_key, expected_order):
    """Test sorting of examples with fully qualified name of a custom sort function."""
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    _check_order(sphinx_app, sort_key, expected_order=expected_order)


def test_collect_gallery_files(tmpdir, gallery_conf):
    """Test that example files are collected properly."""
    rel_filepaths = [
        "examples/file1.py",
        "examples/test.rst",
        "examples/GALLERY_HEADER.rst",
        "examples/folder1/file1.py",
        "examples/folder1/file2.py",
        "examples/folder2/file1.py",
        "tutorials/folder1/file1.py",
        "tutorials/folder2/file1.py",
    ]

    abs_paths = [tmpdir.join(rp) for rp in rel_filepaths]
    for ap in abs_paths:
        ap.ensure()

    examples_path = tmpdir.join("examples")
    dirs = [examples_path.strpath]
    collected_files = set(
        _collect_gallery_files(dirs, gallery_conf, check_filenames=True)
    )
    expected_files = {
        ap.strpath for ap in abs_paths if re.search(r"examples.*\.py$", ap.strpath)
    }

    assert collected_files == expected_files

    tutorials_path = tmpdir.join("tutorials")
    dirs = [examples_path.strpath, tutorials_path.strpath]
    collected_files = set(
        _collect_gallery_files(dirs, gallery_conf, check_filenames=True)
    )
    expected_files = {
        ap.strpath for ap in abs_paths if re.search(r".*\.py$", ap.strpath)
    }

    assert collected_files == expected_files


def test_collect_gallery_files_ignore_pattern(tmpdir, gallery_conf):
    """Test that ignore pattern example files are not collected."""
    rel_filepaths = [
        "examples/file1.py",
        "examples/folder1/fileone.py",
        "examples/folder1/file2.py",
        "examples/folder2/fileone.py",
    ]

    abs_paths = [tmpdir.join(rp) for rp in rel_filepaths]
    for ap in abs_paths:
        ap.ensure()

    gallery_conf["ignore_pattern"] = r"one"
    examples_path = tmpdir.join("examples")
    dirs = [examples_path.strpath]
    collected_files = set(
        _collect_gallery_files(dirs, gallery_conf, check_filenames=True)
    )
    expected_files = {
        ap.strpath
        for ap in abs_paths
        if re.search(r"one", Path(ap.strpath).name) is None
    }

    assert collected_files == expected_files


@pytest.mark.add_conf(
    content=r"""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'copyfile_regex': r'.*\.rst',
}"""
)
@pytest.mark.add_rst(file="own index.rst")
def test_own_index_first(sphinx_app_wrapper):
    """Test `generate_gallery_rst` works when own index gallery is first (and only)."""
    # Issue #1382
    sphinx_app_wrapper.build_sphinx_app()


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'backreferences_dir' : os.path.join('modules', 'gen'),
    'examples_dirs': 'src',
    'gallery_dirs': ['ex'],
    'binder': {'binderhub_url': 'http://test1.com', 'org': 'org',
               'repo': 'repo', 'branch': 'branch',
               'notebooks_dir': 'ntbk_folder',
               'dependencies': 'requirements.txt'}
}"""
)
def test_binder_copy_files(sphinx_app_wrapper):
    """Test that notebooks are copied properly."""
    from sphinx_gallery.interactive_example import copy_binder_files

    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    gallery_conf = sphinx_app.config.sphinx_gallery_conf
    # Create requirements file
    with open(Path(sphinx_app.srcdir, "requirements.txt"), "w"):
        pass
    copy_binder_files(sphinx_app, None)

    for i_file in ["plot_1", "plot_2", "plot_3"]:
        assert Path(
            sphinx_app.outdir,
            "ntbk_folder",
            gallery_conf["gallery_dirs"][0],
            i_file + ".ipynb",
        ).exists()


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
def test_failing_examples_raise_exception(sphinx_app_wrapper):
    example_dir = Path(sphinx_app_wrapper.srcdir, "src")
    bad_line = "print(f'{a[}')"  # never closed bracket inside print -> SyntaxError
    bad_code = f"""\
'''
Failing example
---------------
Should emit a syntax error in the second code block.
'''
1 + 2

# %%
# More

{bad_line}
"""
    bad_line_no = bad_code.split("\n").index(bad_line) + 1
    with open(Path(example_dir, "plot_3.py"), "w", encoding="utf-8") as fid:
        fid.write(bad_code)
    with pytest.raises(ExtensionError) as excinfo:
        sphinx_app_wrapper.build_sphinx_app()
    tb = str(excinfo.value)
    assert "Unexpected failing examples" in tb
    # Check traceback points to correct line (see #1301)
    assert f"line {bad_line_no}" in tb
    assert bad_line in tb


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'filename_pattern': 'plot_1.py',
}"""
)
def test_expected_failing_examples_were_executed(sphinx_app_wrapper):
    """Testing that no exception is issued when broken example is not built.

    See #335 for more details.
    """
    sphinx_app_wrapper.build_sphinx_app()


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'only_warn_on_example_error': True,
}"""
)
def test_only_warn_on_example_error(sphinx_app_wrapper):
    """Test behaviour of only_warn_on_example_error flag."""
    example_dir = Path(sphinx_app_wrapper.srcdir) / "src"
    with open(example_dir / "plot_3.py", "w", encoding="utf-8") as fid:
        fid.write(f"{MINIMAL_HEADER}raise ValueError")
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()

    build_warn = _escape_ansi(sphinx_app._warning.getvalue())
    assert "plot_3.py unexpectedly failed to execute correctly" in build_warn
    assert "WARNING: Here is a summary of the problems" in build_warn


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'only_warn_on_example_error': True,
}"""
)
def test_only_warn_on_example_error_sphinx_warning(sphinx_app_wrapper):
    """Test behaviour of only_warn_on_example_error flag."""
    # https://github.com/sphinx-doc/sphinx/pull/12743/files
    for key in ("warningiserror", "exception_on_warning"):
        if key in inspect.getfullargspec(Sphinx).args:
            sphinx_app_wrapper.kwargs[key] = True
    example_dir = Path(sphinx_app_wrapper.srcdir) / "src"
    with open(example_dir / "plot_3.py", "w", encoding="utf-8") as fid:
        fid.write(f"{MINIMAL_HEADER}raise ValueError")
    with pytest.raises(SphinxWarning) as excinfo:
        sphinx_app_wrapper.build_sphinx_app()
    exc = _escape_ansi(str(excinfo.value))
    assert "plot_3.py unexpectedly failed to execute" in exc


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'expected_failing_examples' :['src/plot_2.py'],
}"""
)
def test_examples_not_expected_to_pass(sphinx_app_wrapper):
    with pytest.raises(ExtensionError) as excinfo:
        sphinx_app_wrapper.build_sphinx_app()
    exc = _escape_ansi(str(excinfo.value))
    assert "expected to fail, but not failing" in exc


@pytest.mark.add_conf(
    content="""
from sphinx_gallery.gen_rst import _sg_call_memory_noop

sphinx_gallery_conf = {
    'show_memory': _sg_call_memory_noop,
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
def test_show_memory_callable(sphinx_app_wrapper):
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()
    status = sphinx_app._status.getvalue()
    assert "0.0 MB" in status, status


@pytest.mark.parametrize(
    "",
    [
        pytest.param(
            id="first notebook cell",
            marks=pytest.mark.add_conf(
                content="""sphinx_gallery_conf = {'first_notebook_cell': 2,}"""
            ),
        ),
        pytest.param(
            id="last notebook cell",
            marks=pytest.mark.add_conf(
                content="""sphinx_gallery_conf = {'last_notebook_cell': 2,}"""
            ),
        ),
    ],
)
def test_notebook_cell_config(sphinx_app_wrapper):
    """Tests that first and last cell configuration validated."""
    with pytest.raises(ConfigError):
        app = sphinx_app_wrapper.create_sphinx_app()
        fill_gallery_conf_defaults(app, app.config, check_keys=False)


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'backreferences_dir': False,
}"""
)
def test_backreferences_dir_config(sphinx_app_wrapper):
    """Tests 'backreferences_dir' type checking."""
    with pytest.raises(ConfigError, match="'backreferences_dir' config allowed types"):
        app = sphinx_app_wrapper.create_sphinx_app()
        fill_gallery_conf_defaults(app, app.config, check_keys=False)


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
def test_minigallery_no_backreferences_dir(sphinx_app_wrapper):
    """Check warning when no backreferences_dir set but minigallery directive used."""
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()
    build_warn = sphinx_app._warning.getvalue()
    assert "'backreferences_dir' config is None, minigallery" in build_warn


@pytest.mark.add_conf(
    content="""
import pathlib

sphinx_gallery_conf = {
    'backreferences_dir': pathlib.Path('.'),
}"""
)
def test_backreferences_dir_pathlib_config(sphinx_app_wrapper):
    """Tests pathlib.Path does not raise exception."""
    app = sphinx_app_wrapper.create_sphinx_app()
    fill_gallery_conf_defaults(app, app.config, check_keys=False)


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
@pytest.mark.add_rst(
    file="""
Header
======

.. minigallery:: index.rst
"""
)
def test_minigallery_not_in_examples_dirs(sphinx_app_wrapper):
    """Check error when minigallery directive's path input not in `examples_dirs`."""
    msg = "minigallery directive error: path input 'index.rst'"
    with pytest.raises(ExtensionError, match=msg):
        sphinx_app_wrapper.build_sphinx_app()


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': ['src', 'src/sub_folder/sub_sub_folder'],
    'gallery_dirs': ['ex', 'ex/sub_folder/sub_sub_folder'],
}"""
)
@pytest.mark.add_rst(
    file="""
Header
======

.. minigallery:: src/sub_folder/sub_sub_folder/plot_nested.py
"""
)
def test_minigallery_multi_match(sphinx_app_wrapper):
    """Check minigallery directive's path input resolution in nested `examples_dirs`.

    When a examples gallery is nested inside another examples gallery, path inputs
    from the nested gallery should resolve to the nested gallery.
    """
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()
    minigallery_html = Path(sphinx_app.outdir) / "minigallery_test.html"
    with open(minigallery_html, "r") as fid:
        mg_html = fid.read()
    # Check thumbnail correct
    assert "_images/sphx_glr_plot_nested_thumb.png" in mg_html
    # Check href correct
    assert "sphx-glr-ex-sub-folder-sub-sub-folder-plot-nested-py" in mg_html


def _get_minigallery_thumbnails(rst_fname):
    """Check the minigallery example file number present in a rst file.

    Note this should only be used for examples in the root
    `sphinx_gallery/tests/testconfs` directory.
    """
    locator = "sphx-glr-thumbcontainer"
    regex = r".+sphx-glr-thumbcontainer.+sphx_glr_plot_(\d)_thumb.+"
    example_numbers = list()
    with open(rst_fname, "r", encoding="utf-8") as fid:
        for line in fid:
            if locator in line:
                example_numbers.append(re.match(regex, line).group(1))
    return example_numbers


@pytest.mark.add_conf(
    content="""
from sphinx_gallery.sorting import FunctionSortKey
from sphinx_gallery.utils import custom_minigallery_sort_order_sorter

sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'minigallery_sort_order': FunctionSortKey(custom_minigallery_sort_order_sorter),
}"""
)
@pytest.mark.add_rst(
    file="""
Header
======

.. minigallery:: src/plot_1.py src/plot_2.py src/plot_3.py
"""
)
def test_minigallery_sort_order_callable(sphinx_app_wrapper):
    """Check `minigallery_sort_order` works when a callable."""
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()

    rst_fname = Path(sphinx_app.outdir, "minigallery_test.html")
    file_numbers = _get_minigallery_thumbnails(rst_fname)
    assert file_numbers == ["3", "2", "1"]


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
}"""
)
@pytest.mark.add_rst(
    file="""
Header
======

.. minigallery:: src/plot_1.py src/plot_1.py
"""
)
def test_minigallery_duplicate_path_input(sphinx_app_wrapper):
    """Check minigallery duplicate input paths are de-deuplicated."""
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()

    rst_fname = Path(sphinx_app.outdir, "minigallery_test.html")
    file_numbers = _get_minigallery_thumbnails(rst_fname)
    assert file_numbers == ["1"]


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'examples_dirs': 'src',
    'gallery_dirs': 'ex',
    'backreferences_dir': 'gen_modules/backreferences',
    'doc_module': ('numpy',),
}"""
)
@pytest.mark.add_rst(
    file="""
Header
======

.. minigallery:: sphinx_gallery.py_source_parser.Block src/plot_1.py
"""
)
def test_minigallery_duplicate_object_path_input(sphinx_app_wrapper):
    """Check object and path input de-deplication works in minigallery directive.

    `Block` is used in `src/plot_1.py`.
    """
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()

    rst_fname = Path(sphinx_app.outdir, "minigallery_test.html")
    file_numbers = _get_minigallery_thumbnails(rst_fname)
    assert file_numbers == ["1"]


def test_write_computation_times_noop(sphinx_app_wrapper):
    app = sphinx_app_wrapper.create_sphinx_app()
    write_computation_times(app.config.sphinx_gallery_conf, None, [])


def test_write_api_usage_noop(sphinx_app_wrapper):
    write_api_entry_usage(sphinx_app_wrapper.create_sphinx_app(), list(), None)


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'pypandoc': ['list',],
}"""
)
def test_pypandoc_config_list(sphinx_app_wrapper):
    """Tests 'pypandoc' type checking."""
    with pytest.raises(
        ConfigError,
        match=r"'pypandoc' config allowed types: \['dict', 'bool'\].",
    ):
        app = sphinx_app_wrapper.create_sphinx_app()
        fill_gallery_conf_defaults(app, app.config, check_keys=False)


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'pypandoc': {'bad key': 1},
}"""
)
def test_pypandoc_config_keys(sphinx_app_wrapper):
    """Tests 'pypandoc' dictionary key checking."""
    with pytest.raises(
        ConfigError, match="'pypandoc' only accepts the following key values:"
    ):
        app = sphinx_app_wrapper.create_sphinx_app()
        fill_gallery_conf_defaults(app, app.config, check_keys=False)


@pytest.mark.add_conf(
    content="""
extensions += ['jupyterlite_sphinx']

sphinx_gallery_conf = {
    'backreferences_dir' : os.path.join('modules', 'gen'),
    'examples_dirs': 'src',
    'gallery_dirs': ['ex'],
}"""
)
def test_create_jupyterlite_contents(sphinx_app_wrapper):
    """Test that JupyterLite contents are created properly."""
    pytest.importorskip("jupyterlite_sphinx")
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    gallery_conf = sphinx_app.config.sphinx_gallery_conf

    create_jupyterlite_contents(sphinx_app, exception=None)

    for i_file in ["plot_1", "plot_2", "plot_3"]:
        assert Path(
            sphinx_app.srcdir,
            "jupyterlite_contents",
            gallery_conf["gallery_dirs"][0],
            i_file + ".ipynb",
        ).exists()


@pytest.mark.add_conf(
    content="""
extensions += ['jupyterlite_sphinx']

sphinx_gallery_conf = {
    'backreferences_dir' : os.path.join('modules', 'gen'),
    'examples_dirs': 'src',
    'gallery_dirs': ['ex'],
    'jupyterlite': {'jupyterlite_contents': 'this_is_the_contents_dir'}
}"""
)
def test_create_jupyterlite_contents_non_default_contents(sphinx_app_wrapper):
    """Test that JupyterLite contents are created properly."""
    pytest.importorskip("jupyterlite_sphinx")
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    gallery_conf = sphinx_app.config.sphinx_gallery_conf

    create_jupyterlite_contents(sphinx_app, exception=None)

    for i_file in ["plot_1", "plot_2", "plot_3"]:
        assert Path(
            sphinx_app.srcdir,
            "this_is_the_contents_dir",
            gallery_conf["gallery_dirs"][0],
            i_file + ".ipynb",
        ).exists()


@pytest.mark.add_conf(
    content="""
sphinx_gallery_conf = {
    'backreferences_dir' : os.path.join('modules', 'gen'),
    'examples_dirs': 'src',
    'gallery_dirs': ['ex'],
}"""
)
def test_create_jupyterlite_contents_without_jupyterlite_sphinx_loaded(
    sphinx_app_wrapper,
):
    """Test JupyterLite contents creation without jupyterlite_sphinx loaded."""
    pytest.importorskip("jupyterlite_sphinx")
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()

    create_jupyterlite_contents(sphinx_app, exception=None)
    assert not Path(sphinx_app.srcdir, "jupyterlite_contents").exists()


@pytest.mark.add_conf(
    content="""
extensions += ['jupyterlite_sphinx']

sphinx_gallery_conf = {
    'backreferences_dir' : os.path.join('modules', 'gen'),
    'examples_dirs': 'src',
    'gallery_dirs': ['ex'],
    'jupyterlite': None,
}"""
)
def test_create_jupyterlite_contents_with_jupyterlite_disabled_via_config(
    sphinx_app_wrapper,
):
    """Test JupyterLite contents created with jupyterlite_sphinx loaded but disabled.

    JupyterLite disabled via config.
    """
    pytest.importorskip("jupyterlite_sphinx")
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()

    create_jupyterlite_contents(sphinx_app, exception=None)
    assert not Path(sphinx_app.outdir, "jupyterlite_contents").exists()


@pytest.mark.add_conf(
    content="""
extensions += ['jupyterlite_sphinx']

def notebook_modification_function(notebook_content, notebook_filename):
    source = f'JupyterLite-specific change for {notebook_filename}'
    markdown_cell = {
        'cell_type': 'markdown',
        'metadata': {},
        'source': source
    }
    notebook_content['cells'] = [markdown_cell] + notebook_content['cells']


sphinx_gallery_conf = {
    'backreferences_dir' : os.path.join('modules', 'gen'),
    'examples_dirs': 'src',
    'gallery_dirs': ['ex'],
    'jupyterlite': {
        'notebook_modification_function': notebook_modification_function
    }
}"""
)
def test_create_jupyterlite_contents_with_modification(sphinx_app_wrapper):
    pytest.importorskip("jupyterlite_sphinx")
    sphinx_app = sphinx_app_wrapper.create_sphinx_app()
    gallery_conf = sphinx_app.config.sphinx_gallery_conf

    create_jupyterlite_contents(sphinx_app, exception=None)

    for i_file in ["plot_1", "plot_2", "plot_3"]:
        notebook_filename = Path(
            sphinx_app.srcdir,
            "jupyterlite_contents",
            gallery_conf["gallery_dirs"][0],
            i_file + ".ipynb",
        )
        assert notebook_filename.exists()

        with open(notebook_filename) as f:
            notebook_content = json.load(f)

        first_cell = notebook_content["cells"][0]
        assert first_cell["cell_type"] == "markdown"
        assert (
            f"JupyterLite-specific change for {notebook_filename}"
            in first_cell["source"]
        )



"""
Failing example test
====================
Test failing thumbnail appears for files in expected_failing_examples.
"""

import matplotlib.pyplot as plt
import numpy as np

plt.pcolormesh(np.random.randn(100, 100))

# %%
# will raise AssertionError

assert False



"""
Choosing the thumbnail figure
=============================

This example demonstrates how to choose the figure that is displayed as the
thumbnail, if the example generates more than one figure. This is done by
specifying the keyword-value pair
``sphinx_gallery_thumbnail_number = <fig number>`` as a
comment somewhere below the docstring in the example file. In this example, we
specify that we wish for the second figure to be the thumbnail.
"""

# Code source: Óscar Nájera
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np


def main():
    """Plot expoential functions."""
    x = np.linspace(-1, 2, 100)
    y = np.exp(x)

    plt.figure()
    plt.plot(x, y)
    plt.xlabel("$x$")
    plt.ylabel(r"$\exp(x)$")

    # The next line sets the thumbnail for the second figure in the gallery
    # (plot with negative exponential in orange)
    # sphinx_gallery_thumbnail_number = 2
    plt.figure()
    plt.plot(x, -np.exp(-x), color="orange", linewidth=4)
    plt.xlabel("$x$")
    plt.ylabel(r"$-\exp(-x)$")
    # To avoid matplotlib text output
    plt.show()


if __name__ == "__main__":
    main()



# Author: Óscar Nájera
# License: 3-clause BSD
r"""Test source parser."""

import os.path as op
import textwrap

import pytest
from sphinx.errors import ExtensionError

import sphinx_gallery.py_source_parser as sg


def test_get_docstring_and_rest(unicode_sample, tmpdir, monkeypatch):
    """Test `_get_docstring_and_rest` correctly splits docstring and rest of example."""
    docstring, rest, lineno, _ = sg._get_docstring_and_rest(unicode_sample)
    assert "Únicode" in docstring
    assert "heiß" in rest
    # degenerate
    fname = op.join(str(tmpdir), "temp")
    with open(fname, "w") as fid:
        fid.write('print("hello")\n')
    with pytest.raises(ExtensionError, match="Could not find docstring"):
        sg._get_docstring_and_rest(fname)
    with open(fname, "w") as fid:
        fid.write("print hello\n")
    assert sg._get_docstring_and_rest(fname)[0] == sg.SYNTAX_ERROR_DOCSTRING
    monkeypatch.setattr(sg, "parse_source_file", lambda x: ("", None))
    with pytest.raises(ExtensionError, match="only supports modules"):
        sg._get_docstring_and_rest("")


@pytest.mark.parametrize(
    "content, file_conf",
    [
        ("No config\nin here.", {}),
        ("# sphinx_gallery_line_numbers = True", {"line_numbers": True}),
        ("  #   sphinx_gallery_line_numbers   =   True   ", {"line_numbers": True}),
        ("#sphinx_gallery_line_numbers=True", {"line_numbers": True}),
        ("#sphinx_gallery_thumbnail_number\n=\n5", {"thumbnail_number": 5}),
        ("#sphinx_gallery_thumbnail_number=1foo", None),
        ("# sphinx_gallery_defer_figures", {}),
    ],
)
def test_extract_file_config(content, file_conf, log_collector):
    """Test file config correctly extracted."""
    if file_conf is None:
        assert sg.extract_file_config(content) == {}
        log_collector.warning.assert_called_once()
        assert "1foo" == log_collector.warning.call_args[0][2]
    else:
        assert sg.extract_file_config(content) == file_conf
        log_collector.warning.assert_not_called()


@pytest.mark.parametrize(
    "contents, result",
    [
        ("No config\nin here.", "No config\nin here."),
        ("# sphinx_gallery_line_numbers = True", ""),
        ("  #   sphinx_gallery_line_numbers   =   True   ", ""),
        ("#sphinx_gallery_line_numbers=True", ""),
        ("#sphinx_gallery_thumbnail_number\n=\n5", ""),
        ("a = 1\n# sphinx_gallery_line_numbers = True\nb = 1", "a = 1\nb = 1"),
        ("a = 1\n\n# sphinx_gallery_line_numbers = True\n\nb = 1", "a = 1\n\n\nb = 1"),
        (
            "# comment\n# sphinx_gallery_line_numbers = True\n# comment 2",
            "# comment\n# comment 2",
        ),
        ("# sphinx_gallery_defer_figures", ""),
    ],
)
def test_remove_config_comments(contents, result):
    assert sg.remove_config_comments(contents) == result


def test_remove_ignore_comments():
    normal_code = "# Regular code\n# should\n# be untouched!"
    assert sg.remove_ignore_blocks(normal_code) == normal_code

    mismatched_code = "x=5\n# sphinx_gallery_start_ignore\ny=4"
    with pytest.raises(ExtensionError) as error:
        sg.remove_ignore_blocks(mismatched_code)
    assert "must have a matching" in str(error)

    code_with_ignores = textwrap.dedent(
        """\
    # Indented ignores should work
        # sphinx_gallery_start_ignore
        # The variable name should do nothing
        sphinx_gallery_end_ignore = 0
        # sphinx_gallery_end_ignore

    # New line above should stay intact
    # sphinx_gallery_start_ignore
    # sphinx_gallery_end_ignore
    # Empty ignore blocks are fine too
    """
    )
    assert sg.remove_ignore_blocks(code_with_ignores) == textwrap.dedent(
        """\
    # Indented ignores should work

    # New line above should stay intact
    # Empty ignore blocks are fine too
    """
    )



"""
Force images to be displayed on separate lines per block
========================================================
This example demonstrates how the ``sphinx_gallery_multi_image_block`` option can be
used to force images to be displayed on separate lines for a specific block, rather than
the default behaviour of displaying them side-by-side.
"""

import matplotlib.pyplot as plt
import numpy as np

# %%

# Default behaviour
fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))


# %%

# sphinx_gallery_multi_image_block = "single"

# Non-default behaviour for just this cell
fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))



"""Testing image scrapers."""

import os

import pytest
from sphinx.errors import ConfigError, ExtensionError

import sphinx_gallery
from sphinx_gallery.gen_gallery import _fill_gallery_conf_defaults
from sphinx_gallery.scrapers import (
    _KNOWN_IMG_EXTS,
    SG_IMAGE,
    ImagePathIterator,
    _reset_matplotlib,
    figure_rst,
    matplotlib_scraper,
    save_figures,
)


@pytest.fixture(scope="function")
def make_gallery_conf(tmpdir):
    """Sets up a test sphinx-gallery configuration."""
    # Skip if numpy not installed
    pytest.importorskip("numpy")

    def make_gallery_conf(init=None):
        gallery_conf = _fill_gallery_conf_defaults(init or {})
        gallery_conf.update(
            src_dir=str(tmpdir), examples_dir=str(tmpdir), gallery_dir=str(tmpdir)
        )

        return gallery_conf

    return make_gallery_conf


class matplotlib_svg_scraper:
    """Test matplotlib svg scraper."""

    def __repr__(self):
        return self.__class__.__name__

    def __call__(self, *args, **kwargs):
        """Call matplotlib scraper with 'svg' format."""
        return matplotlib_scraper(*args, format="svg", **kwargs)


@pytest.mark.parametrize("ext", ("png", "svg"))
def test_save_matplotlib_figures(make_gallery_conf, ext):
    """Test matplotlib figure save."""
    gallery_conf = make_gallery_conf(
        {"image_scrapers": (matplotlib_svg_scraper(),)} if ext == "svg" else {}
    )
    import matplotlib.pyplot as plt  # nest these so that Agg can be set

    plt.plot(1, 1)
    fname_template = os.path.join(gallery_conf["gallery_dir"], "image{0}.png")
    image_path_iterator = ImagePathIterator(fname_template)
    block = ("",) * 3
    block_vars = dict(image_path_iterator=image_path_iterator)
    image_rst = save_figures(block, block_vars, gallery_conf)
    assert len(image_path_iterator) == 1
    fname = f"/image1.{ext}"
    assert fname in image_rst
    fname = gallery_conf["gallery_dir"] + fname
    assert os.path.isfile(fname)

    def _create_two_images():
        image_path_iterator.next()
        image_path_iterator.next()
        plt.plot(1, 1)
        plt.figure()
        plt.plot(1, 1)

    # Test capturing 2 images with shifted start number
    _create_two_images()
    image_rst = save_figures(block, block_vars, gallery_conf)
    assert len(image_path_iterator) == 5
    for ii in range(4, 6):
        fname = f"/image{ii}.{ext}"
        assert fname in image_rst
        fname = gallery_conf["gallery_dir"] + fname
        assert os.path.isfile(fname)

    # Test `sphinx_gallery_multi_image(_block)` variables work; these variables prevent
    # images with `sphx-glr-single-img` classes from being converted to
    # `sphx-glr-multi-img` classes; requires > 1 image

    # Test file-wide `sphinx_gallery_multi_image` variable
    _create_two_images()
    block_vars["file_conf"] = {"multi_image": "single"}
    image_rst = save_figures(block, block_vars, gallery_conf)
    assert "sphx-glr-single-img" in image_rst
    assert "sphx-glr-multi-img" not in image_rst

    # Test block-specific `sphinx_gallery_multi_image_block` variable
    # (test with default `sphinx_gallery_multi_image`, i.e. != "single")
    _create_two_images()
    block_vars["file_conf"] = {}
    block_vars["multi_image"] = "single"
    image_rst = save_figures(block, block_vars, gallery_conf)
    assert "sphx-glr-single-img" in image_rst
    assert "sphx-glr-multi-img" not in image_rst
    # (test block-specific setting overrides file-wide setting)
    _create_two_images()
    block_vars["file_conf"] = {"multi_image": "single"}
    block_vars["multi_image"] = "multi"
    image_rst = save_figures(block, block_vars, gallery_conf)
    assert "sphx-glr-single-img" not in image_rst
    assert "sphx-glr-multi-img" in image_rst


def test_image_srcset_config(make_gallery_conf):
    with pytest.raises(ConfigError, match="'image_srcset' config allowed"):
        make_gallery_conf({"image_srcset": "2x"})
    with pytest.raises(ConfigError, match="Invalid value for image_srcset parameter"):
        make_gallery_conf({"image_srcset": [False]})
    with pytest.raises(ConfigError, match="Invalid value for image_srcset parameter"):
        make_gallery_conf({"image_srcset": ["200"]})

    conf = make_gallery_conf({"image_srcset": ["2x"]})
    assert conf["image_srcset"] == [2.0]
    conf = make_gallery_conf({"image_srcset": ["1x", "2x"]})
    assert conf["image_srcset"] == [2.0]  # "1x" is implied.


def test_save_matplotlib_figures_hidpi(make_gallery_conf):
    """Test matplotlib hidpi figure save."""
    gallery_conf = make_gallery_conf({"image_srcset": ["2x"]})
    ext = "png"

    import matplotlib.pyplot as plt  # nest these so that Agg can be set

    plt.plot(1, 1)
    fname_template = os.path.join(gallery_conf["gallery_dir"], "image{0}.png")
    image_path_iterator = ImagePathIterator(fname_template)
    block = ("",) * 3
    block_vars = dict(image_path_iterator=image_path_iterator)
    image_rst = save_figures(block, block_vars, gallery_conf)

    fname = f"/image1.{ext}"
    assert fname in image_rst
    assert f"/image1_2_00x.{ext} 2.00x" in image_rst

    assert len(image_path_iterator) == 1
    fname = gallery_conf["gallery_dir"] + fname
    fnamehi = gallery_conf["gallery_dir"] + f"/image1_2_00x.{ext}"

    assert os.path.isfile(fname)
    assert os.path.isfile(fnamehi)

    # Test capturing 2 images with shifted start number
    image_path_iterator.next()
    image_path_iterator.next()
    plt.plot(1, 1)
    plt.figure()
    plt.plot(1, 1)
    image_rst = save_figures(block, block_vars, gallery_conf)
    assert len(image_path_iterator) == 5
    for ii in range(4, 6):
        fname = f"/image{ii}.{ext}"
        assert fname in image_rst

        fname = gallery_conf["gallery_dir"] + fname
        assert os.path.isfile(fname)
        fname = f"/image{ii}_2_00x.{ext}"
        assert fname in image_rst
        fname = gallery_conf["gallery_dir"] + fname
        assert os.path.isfile(fname)


def _custom_func(x, y, z):
    return y["image_path_iterator"].next()


def test_custom_scraper(make_gallery_conf, monkeypatch):
    """Test custom scrapers."""
    # Test the API contract for custom scrapers
    with monkeypatch.context() as m:
        m.setattr(
            sphinx_gallery, "_get_sg_image_scraper", lambda: _custom_func, raising=False
        )
        for cust in (_custom_func, "sphinx_gallery"):
            # smoke test that it works
            make_gallery_conf({"image_scrapers": [cust]})
    # degenerate
    # without the monkey patch to add sphinx_gallery._get_sg_image_scraper,
    # we should get an error
    with pytest.raises(ConfigError, match="Unknown string option"):
        make_gallery_conf({"image_scrapers": ["sphinx_gallery"]})

    # other degenerate conditions
    with pytest.raises(ConfigError, match="Unknown string option for image_scraper"):
        make_gallery_conf({"image_scrapers": ["foo"]})
    for cust, msg in [
        (_custom_func, "did not produce expected image"),
        (lambda x, y, z: 1.0, "was not a string"),
    ]:
        conf = make_gallery_conf({"image_scrapers": [cust]})
        fname_template = os.path.join(conf["gallery_dir"], "image{0}.png")
        image_path_iterator = ImagePathIterator(fname_template)
        block = ("",) * 3
        block_vars = dict(image_path_iterator=image_path_iterator)
        with pytest.raises(ExtensionError, match=msg):
            save_figures(block, block_vars, conf)
    # degenerate string interface
    with monkeypatch.context() as m:
        m.setattr(sphinx_gallery, "_get_sg_image_scraper", "foo", raising=False)
        with pytest.raises(ConfigError, match="^Unknown string option for image_"):
            make_gallery_conf({"image_scrapers": ["sphinx_gallery"]})
    with monkeypatch.context() as m:
        m.setattr(sphinx_gallery, "_get_sg_image_scraper", lambda: "foo", raising=False)
        with pytest.raises(ConfigError, match="craper.*must be callable"):
            make_gallery_conf({"image_scrapers": ["sphinx_gallery"]})


@pytest.mark.parametrize("ext", _KNOWN_IMG_EXTS)
def test_figure_rst(ext):
    """Test reST generation of images."""
    figure_list = ["sphx_glr_plot_1." + ext]
    image_rst = figure_rst(figure_list, ".")
    single_image = f"""
.. image-sg:: /sphx_glr_plot_1.{ext}
   :alt: pl
   :srcset: /sphx_glr_plot_1.{ext}
   :class: sphx-glr-single-img
"""
    assert image_rst == single_image

    image_rst = figure_rst(figure_list + ["second." + ext], ".")

    image_list_rst = f"""
.. rst-class:: sphx-glr-horizontal


    *

      .. image-sg:: /sphx_glr_plot_1.{ext}
          :alt: pl
          :srcset: /sphx_glr_plot_1.{ext}
          :class: sphx-glr-multi-img

    *

      .. image-sg:: /second.{ext}
          :alt: pl
          :srcset: /second.{ext}
          :class: sphx-glr-multi-img
"""
    assert image_rst == image_list_rst


def test_figure_rst_path():
    """Test figure path correct in figure reSt."""
    # Tests issue #229
    local_img = [os.path.join(os.getcwd(), "third.png")]
    image_rst = figure_rst(local_img, ".")

    single_image = SG_IMAGE % ("third.png", "", "/third.png")
    assert image_rst == single_image


def test_figure_rst_srcset():
    """Test reST generation of images with srcset paths correct."""
    figure_list = ["sphx_glr_plot_1.png"]
    hipaths = [{0: "sphx_glr_plot_1.png", 2.0: "sphx_glr_plot_1_2_00.png"}]
    image_rst = figure_rst(figure_list, ".", srcsetpaths=hipaths)
    single_image = """
.. image-sg:: /sphx_glr_plot_1.png
   :alt: pl
   :srcset: /sphx_glr_plot_1.png, /sphx_glr_plot_1_2_00.png 2.00x
   :class: sphx-glr-single-img
"""
    assert image_rst == single_image

    hipaths += [{0: "second.png", 2.0: "second_2_00.png"}]
    image_rst = figure_rst(figure_list + ["second.png"], ".", srcsetpaths=hipaths)
    image_list_rst = """
.. rst-class:: sphx-glr-horizontal


    *

      .. image-sg:: /sphx_glr_plot_1.png
          :alt: pl
          :srcset: /sphx_glr_plot_1.png, /sphx_glr_plot_1_2_00.png 2.00x
          :class: sphx-glr-multi-img

    *

      .. image-sg:: /second.png
          :alt: pl
          :srcset: /second.png, /second_2_00.png 2.00x
          :class: sphx-glr-multi-img
"""
    assert image_rst == image_list_rst

    # test issue #229
    local_img = [os.path.join(os.getcwd(), "third.png")]
    image_rst = figure_rst(local_img, ".")

    single_image = SG_IMAGE % ("third.png", "", "/third.png")
    assert image_rst == single_image


def test_iterator():
    """Test ImagePathIterator."""
    ipi = ImagePathIterator("foo{0}")
    ipi._stop = 10
    with pytest.raises(ExtensionError, match="10 images"):
        for ii in ipi:
            pass


def test_reset_matplotlib(make_gallery_conf):
    """Test _reset_matplotlib."""
    import matplotlib

    matplotlib.rcParams["lines.linewidth"] = 42
    matplotlib.units.registry.clear()

    gallery_conf = make_gallery_conf()
    _reset_matplotlib(gallery_conf, "")

    assert matplotlib.rcParams["lines.linewidth"] != 42
    assert len(matplotlib.units.registry) > 0



"""
Force plots to be displayed on separate lines
=============================================
This example demonstrates how the visualisation of multiple plots produced from a single
code block can be controlled. The default behaviour is to stack plots side-by-side,
however this can be overridden to display each plot created by the code block on a
separate line, preserving their size.

There are two config options to control this behaviour:

- a file-wide ``sphinx_gallery_multi_image`` variable
- a code block-specific ``sphinx_gallery_multi_image_block`` variable

Setting these variables to ``"single"`` will force plots to be displayed on separate
lines. Default behaviour is to treat these variables as being set to ``"multi"``.

Below we demonstrate how the file-wide ``sphinx_gallery_multi_image`` variable can be
used to display plots on separate lines.
"""

# Code source: Thomas S. Binns
# License: BSD 3 clause

# sphinx_gallery_multi_image = "single"

import matplotlib.pyplot as plt
import numpy as np

# %%

# Plots will be shown on separate lines

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))

########################################################################################
# Now, we show how the ``sphinx_gallery_multi_image_block`` variable can be used to
# control the behaviour for a specific code block, here reverting to the default
# behaviour of stacking plots side-by-side.

# %%

# sphinx_gallery_multi_image_block = "multi"
# ↑↑↑ Return to default behaviour for just this cell

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))

fig, ax = plt.subplots(1, 1, figsize=(8, 4))
ax.pcolormesh(np.random.randn(100, 100))



# Author: Óscar Nájera
# License: 3-clause BSD
"""Testing the rst files generator."""

import sys
from unittest.mock import MagicMock, patch

import pytest
from sphinx.errors import ExtensionError

import sphinx_gallery.backreferences as sg
from sphinx_gallery.gen_rst import _sanitize_rst
from sphinx_gallery.py_source_parser import Block, split_code_and_text_blocks

REFERENCE = r"""
.. raw:: html

    <div class="sphx-glr-thumbcontainer" tooltip="{0}">

.. only:: html

  .. image:: /fake_dir/images/thumb/sphx_glr_test_file_thumb.png
    :alt:

  :ref:`sphx_glr_fake_dir_test_file.py`

.. raw:: html

      <div class="sphx-glr-thumbnail-title">test title</div>
    </div>

{1}"""


@pytest.mark.parametrize(
    "content, tooltip, is_backref",
    [
        # HTML sanitizing
        ('<"test">', "&lt;&quot;test&quot;&gt;", False),
        # backref support
        ("test formatting", "test formatting", True),
        # reST sanitizing
        (
            "1 :class:`~a.b`. 2 :class:`a.b` 3 :ref:`whatever <better name>`",
            "1 b. 2 a.b 3 better name",
            False,
        ),
        ("use :meth:`mne.io.Raw.plot_psd` to", "use mne.io.Raw.plot_psd to", False),
        (
            "`this` and ``that``; and `these things` and ``those things``",
            "this and that; and these things and those things",
            False,
        ),
        (
            "See `.MyClass` and `~.MyClass.close`",
            "See MyClass and close",
            False,
        ),
    ],
)
def test_thumbnail_div(content, tooltip, is_backref):
    """Test if the thumbnail div generates the correct string."""
    with pytest.raises(ExtensionError, match="internal Sphinx-Gallery thumb"):
        html_div = sg._thumbnail_div(
            "fake_dir", "", "test_file.py", '<"test">', '<"title">'
        )
    content = _sanitize_rst(content)
    title = "test title"
    html_div = sg._thumbnail_div(
        "fake_dir",
        "",
        "test_file.py",
        content,
        title,
        is_backref=is_backref,
        check=False,
    )
    if is_backref:
        extra = """
.. only:: not html

 * :ref:`sphx_glr_fake_dir_test_file.py`
"""
    else:
        extra = ""
    reference = REFERENCE.format(tooltip, extra)
    assert html_div == reference


def test_identify_names(unicode_sample, gallery_conf):
    """Test name identification."""
    expected = {
        "os.path.join": [
            {
                "name": "join",
                "module": "os.path",
                "module_short": "os.path",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        "br.identify_names": [
            {
                "name": "identify_names",
                "module": "sphinx_gallery.back_references",
                "module_short": "sphinx_gallery.back_references",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        "identify_names": [
            {
                "name": "identify_names",
                "module": "sphinx_gallery.back_references",
                "module_short": "sphinx_gallery.back_references",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        # Check `get_short_module_name` points to correct object.
        # Here, `matplotlib.pyplot.figure` (func) can be shortened to
        # `matplotlib.figure` (module) (but should not be)
        "plt.figure": [
            {
                "name": "figure",
                "module": "matplotlib.pyplot",
                "module_short": "matplotlib.pyplot",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        # Check we get a in a().b in `visit_Attribute`
        "DummyClass": [
            {
                "name": "DummyClass",
                "module": "sphinx_gallery._dummy",
                "module_short": "sphinx_gallery._dummy",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        "NestedDummyClass": [
            {
                "name": "NestedDummyClass",
                "module": "sphinx_gallery._dummy.nested",
                "module_short": "sphinx_gallery._dummy",
                "is_class": False,
                "is_explicit": False,
            }
        ],
    }
    _, script_blocks = split_code_and_text_blocks(unicode_sample)
    ref_regex = sg._make_ref_regex()
    res = sg.identify_names(script_blocks, ref_regex)
    assert expected == res


@pytest.mark.parametrize(("mock", "short_module"), [("same", "A"), ("diff", "A.B")])
def test_get_short_module_name(mock, short_module):
    """Check `_get_short_module_name` correctly finds shortest module."""
    if mock == "same":
        mock_mod_1 = mock_mod_2 = MagicMock()
    else:
        mock_mod_1 = MagicMock()
        mock_mod_2 = MagicMock()

    # Mock will return whatever object/attr we ask of it
    # When mock objects are the same, the `obj_name` from "A" is the same as
    # `obj_name` from "A.B", so "A" is an accepted short module
    with patch.dict(sys.modules, {"A": mock_mod_1, "A.B": mock_mod_2}):
        short_mod = sg._get_short_module_name("A.B", "C")
        short_mod_with_attr = sg._get_short_module_name("A.B", "C.D")
        assert short_mod == short_mod_with_attr == short_module
        # Deleting desired attr will cause failure of `hasattr` check (we
        # check this to ensure it is the right class)
        del mock_mod_2.C.D
        short_mod_no_attr = sg._get_short_module_name("A.B", "C.D")
        assert short_mod_no_attr is None


def test_identify_names_implicit(tmpdir, gallery_conf):
    """Test implicit name identification."""
    code_str = b"""
'''
Title
-----

This is an example.
'''
# -*- coding: utf-8 -*-
# \xc3\x9f
from a.b import c
import d as e
import h.i
print(c)
e.HelloWorld().f.g
h.i.j()
"""
    expected = {
        "c": [
            {
                "name": "c",
                "module": "a.b",
                "module_short": "a.b",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        "e.HelloWorld": [
            {
                "name": "HelloWorld",
                "module": "d",
                "module_short": "d",
                "is_class": False,
                "is_explicit": False,
            }
        ],
        "h.i.j": [
            {
                "name": "j",
                "module": "h.i",
                "module_short": "h.i",
                "is_class": False,
                "is_explicit": False,
            }
        ],
    }

    fname = tmpdir.join("identify_names.py")
    fname.write(code_str, "wb")

    _, script_blocks = split_code_and_text_blocks(fname.strpath)
    ref_regex = sg._make_ref_regex()
    res = sg.identify_names(script_blocks, ref_regex)

    assert expected == res


cobj = dict(module="m", module_short="m", name="n", is_class=False, is_explicit=True)


@pytest.mark.parametrize(
    "text, default_role, ref, cobj",
    [
        (":func:`m.n`", None, "m.n", cobj),
        (":func:`~m.n`", "obj", "m.n", cobj),
        (":func:`Title <m.n>`", None, "m.n", cobj),
        (":func:`!m.n` or `!t <~m.n>`", None, None, None),
        ("`m.n`", "obj", "m.n", cobj),
        ("`m.n`", None, None, None),  # see comment below
        (":ref:`m.n`", None, None, None),
        ("`m.n`", "ref", None, None),
        ("``literal``", "obj", None, None),
    ],
    ids=[
        "regular",
        "show only last component",
        "with title",
        "no link for !",
        "default_role obj",
        "no default_role",  # see comment below
        "non-python role",
        "non-python default_role",
        "literal",
    ],
)
# the sphinx default value for default_role is None = no change, the docutils
# default is title-reference (set by the default-role directive), see
# www.sphinx-doc.org/en/master/usage/configuration.html#confval-default_role
# and docutils.sourceforge.io/docs/ref/rst/roles.html
def test_identify_names_explicit(text, default_role, ref, cobj, gallery_conf):
    """Test explicit name identification."""
    default_role = "" if default_role is None else default_role
    script_blocks = [Block("text", text, 1)]
    expected = {ref: [cobj]} if ref else {}
    ref_regex = sg._make_ref_regex(default_role)
    actual = sg.identify_names(script_blocks, ref_regex)
    assert expected == actual



"""
Test ``__future__`` imports across cells
----------------------------------------

This example tests that ``__future__`` imports works across cells.
"""

from __future__ import division, print_function

import matplotlib

####################
# Dummy section, with :func:`sphinx_gallery.backreferences.NameFinder` ref.

assert 3 / 2 == 1.5
print(3 / 2, end="")

# testing reset of mpl
orig_dpi = 80.0 if matplotlib.__version__[0] < "2" else 100.0
assert matplotlib.rcParams["figure.dpi"] == orig_dpi
matplotlib.rcParams["figure.dpi"] = 90.0
assert matplotlib.rcParams["figure.dpi"] == 90.0



"""
Example that fails to execute (with normal thumbnail behaviour)
===============================================================

By default, examples with code blocks that raise an error will have the broken
image stamp as their gallery thumbnail. However, this may not be desired, e.g.
if only part of the example is expected to fail and it should not look like the
entire example fails.

In these cases, the `sphinx_gallery_failing_thumbnail` variable can be set to
``False``, which will change the thumbnail selection to the default behaviour
as for non-failing examples.
"""

# Code source: Thomas S. Binns
# License: BSD 3 clause
# sphinx_gallery_line_numbers = True

# sphinx_gallery_failing_thumbnail = False

import matplotlib.pyplot as plt
import numpy as np

plt.pcolormesh(np.random.randn(100, 100))

# %%
# This block will raise an AssertionError

assert False
